"""
Responsible for reflecting BGP-learned routes back to VPP.
"""

import select
import signal
import threading
import time
from enum import IntEnum
from multiprocessing import Pipe

from pyroute2 import IPRoute

from vrouter.client.client import ExtendedVPPAPIClient
from vrouter.common import models
from vrouter.common.utils import get_main_logger

logger = get_main_logger()

# route reflector configuration (must be set from PAPI++ server, but that starts after route_reflector due to
# FRR container starting first for FRR Agent IP address retrieval)
_config_lock = threading.Lock()
_papi_server_host = ""
_papi_server_port = 0


class Proto(IntEnum):  # https://fossies.org/linux/iproute2/etc/iproute2/rt_protos
    UNSPEC = 0
    REDIRECT = 1
    KERNEL = 2
    BOOT = 3
    STATIC = 4
    GATED = 8
    RA = 9
    MRT = 10
    ZEBRA = 11
    BIRD = 12
    DNROUTED = 13
    XORP = 14
    NTK = 15
    DHCP = 16
    KEEPALIVED = 18
    BABEL = 42
    OPENR = 99
    BGP = 186
    ISIS = 187
    OSPF = 188
    RIP = 189
    EIGRP = 192


def configure_route_reflector(
    papi_server_host: str = "0.0.0.0", papi_server_port: int = 9999
):
    with _config_lock:
        global _papi_server_host, _papi_server_port
        _papi_server_host = papi_server_host
        _papi_server_port = papi_server_port


def start_route_reflector_in_background():
    quit_signaling_out, quit_signaling_in = Pipe()
    rr_thread = threading.Thread(target=_route_reflection, args=(quit_signaling_in,))
    rr_thread.start()

    def graceful_shutdown_callback():
        quit_signaling_out.send("")
        rr_thread.join()
        quit_signaling_out.close()
        quit_signaling_in.close()

    return graceful_shutdown_callback


def _wait_for_configuration():
    while True:
        logger.info("waiting for configuration...")
        with _config_lock:
            global _papi_server_host, _papi_server_port
            if _papi_server_host:
                return _papi_server_host, _papi_server_port
        time.sleep(1)


def _route_reflection(quit_signaling_in):
    logger.info("started route reflection thread")
    papi_server_host, papi_server_port = _wait_for_configuration()
    logger.debug("create poll object")
    poll = select.poll()

    logger.debug("create IPRoute object")
    with IPRoute(nlm_generator=True) as ipr:
        logger.debug("register IPRoute for select.POLLIN")
        poll.register(ipr, select.POLLIN)
        poll.register(quit_signaling_in, select.POLLIN)

        logger.debug("bind IPRoute to get broadcasts")
        ipr.bind()

        added_routes = dict()  # dict of previously added routes

        while True:
            logger.debug("waiting for new netlink data")
            try:
                # Note: could use also poll.poll(<timeout in milliseconds>) but that would mean to have compromise
                # between graceful shutdown length and listening interruptions
                notifications = poll.poll()
            except Exception as e:
                logger.exception("exception from netlink data wait", e)
                notifications = []  # nothing was retrieved
            logger.info(f"processing new (poll.poll) notifications: {notifications}")
            if any(
                notification[0] == quit_signaling_in.fileno()
                for notification in notifications
            ):
                logger.info(
                    "detected command for graceful shutdown -> ending route reflection thread"
                )
                break  # something came from quit signaling pipe -> need to end listening to netlink and stop

            if notifications:
                logger.debug("loading ip address replace mapping")
                # quick and dirty solution how to replace ip address in learned routes from BGP/OSPF,
                # it is needed due to using GRE tunnel (LCP hack) that have different ip address to the wireguard
                # tunnel addresses and we want to use wg tunnel addresses for vpp route installation (so that dataplane
                # won't go through GRE tunnel that is LCP hack for controlplane)
                # FIXME this should be configurable from Vrouter client
                ip_address_replace_mapping = _get_ip_address_replace_mapping()
                logger.debug("retrieving netlink data")
                _handle_netlink_messages(
                    ipr.get(),
                    added_routes,
                    papi_server_host,
                    papi_server_port,
                    ip_address_replace_mapping,
                )


def _get_ip_address_replace_mapping():
    ip_address_replace_mapping = {}
    try:
        with open("/rr-ip-replace.txt") as input_file:
            for line in input_file:
                old_ip, new_ip = line.partition("=>")[::2]
                ip_address_replace_mapping[old_ip.strip()] = new_ip.strip()
    except FileNotFoundError:
        logger.debug("input file for ip_address_replace_mapping not present")

    logger.debug("ip_address_replace_mapping: " + str(ip_address_replace_mapping))
    return ip_address_replace_mapping


def _handle_netlink_messages(
    messages,
    added_routes,
    papi_server_host,
    papi_server_port,
    ip_address_replace_mapping,
):
    for message in messages:
        logger.debug(f"got netlink message: {message}")
        if message["event"] in ["RTM_NEWROUTE", "RTM_DELROUTE"] and message[
            "proto"
        ] in [int(Proto.BGP), int(Proto.OSPF)]:
            route = _get_route(message, ip_address_replace_mapping)
            route_key = route.destination_network
            try:
                with ExtendedVPPAPIClient(
                    server_host=papi_server_host, server_port=papi_server_port
                ) as client:
                    if message["event"] == "RTM_NEWROUTE":
                        if added_routes.get(route_key):
                            logger.info(
                                f"deleting previous route due to update: {added_routes.get(route_key)}"
                            )
                            reply = client.delete_configuration(
                                added_routes.get(route_key)
                            )
                            if (
                                not reply.vpp_apply_success
                                or not reply.processing_error == ""
                            ):
                                logger.error(
                                    f"couldn't delete previous route(update) {added_routes.get(route_key)}"
                                )
                            del added_routes[route_key]

                        logger.info(f"adding new route {route}")
                        reply = client.add_configuration(route)
                        if (
                            not reply.vpp_apply_success
                            or not reply.all_added_items_applied_to_vpp
                            or not reply.processing_error == ""
                        ):
                            logger.error(f"couldn't add new route route {route}")
                        added_routes[route_key] = route
                    elif message["event"] == "RTM_DELROUTE":
                        logger.info(f"deleting previous route {route}")
                        reply = client.delete_configuration(route)
                        if (
                            not reply.vpp_apply_success
                            or not reply.processing_error == ""
                        ):
                            logger.error(f"couldn't delete previous route {route}")
                        del added_routes[route_key]
            except Exception:
                logger.exception(
                    f"can't create/delete route from route reflector (route={route})"
                )


def _get_route(netlink_message, ip_address_replace_mapping):
    # extract some common data from netlink message
    logger.debug("extracting info from message")
    dest_net = (
        f"{netlink_message.get_attr('RTA_DST')}/{netlink_message['dst_len']}"
        if netlink_message.get_attr("RTA_DST")
        else "0.0.0.0/0"
    )
    priority = (
        netlink_message.get_attr("RTA_PRIORITY")
        if netlink_message.get_attr("RTA_PRIORITY")
        else 1
    )

    # single output path
    if netlink_message.get_attr("RTA_MULTIPATH") is None:
        gw_addr = (
            netlink_message.get_attr("RTA_GATEWAY")
            if netlink_message.get_attr("RTA_GATEWAY") is not None
            else ""
        )
        out_interface_name = _get_output_interface_name(
            netlink_message.get_attr("RTA_OIF")
        )

        # reflect route change (send route change to PAPI++ server)
        # Requirement: name of interface must be the same as logical name in PAPI++ server
        return models.RouteConfigurationItem(
            destination_network=_ip_address_replace(
                dest_net, ip_address_replace_mapping
            ),
            next_hop_address=_ip_address_replace(gw_addr, ip_address_replace_mapping),
            outgoing_interface=out_interface_name,
            weight=priority,
        )

    # multiple output paths
    route = models.RouteConfigurationItem(
        destination_network=_ip_address_replace(dest_net, ip_address_replace_mapping),
        multi_output_paths=[],
    )
    for multipath in netlink_message.get_attr("RTA_MULTIPATH"):
        gw_addr = (
            multipath.get_attr("RTA_GATEWAY")
            if multipath.get_attr("RTA_GATEWAY") is not None
            else ""
        )
        out_interface_name = _get_output_interface_name(multipath["oif"])

        # reflect route change (send route change to PAPI++ server)
        # Requirement: name of interface must be the same as logical name in PAPI++ server
        route.multi_output_paths.append(
            models.RouteOutputPath(
                next_hop_address=_ip_address_replace(
                    gw_addr, ip_address_replace_mapping
                ),
                outgoing_interface=out_interface_name,
                weight=priority,
            )
        )
    return route


def _ip_address_replace(
    ip_or_network: str, ip_address_replace_mapping: dict[str, str]
) -> str:
    if ip_address_replace_mapping.get(ip_or_network):
        return ip_address_replace_mapping.get(
            ip_or_network
        )  # ip address that needs replacing
    if "/" in ip_or_network:  # network
        split = ip_or_network.split("/", 1)
        ip = split[0]
        mask = split[1]
        if ip_address_replace_mapping.get(ip):
            return (
                ip_address_replace_mapping.get(ip) + "/" + mask
            )  # network that needs its ip part to be replaced
    return ip_or_network


def _get_output_interface_name(netlink_oif_id):
    out_interface_name = ""  # no output interface
    if netlink_oif_id is not None:
        try:
            with IPRoute(
                nlm_generator=True
            ) as ipr2:  # if using IPRoute used for notification, it gets stucked
                link = ipr2.get_links(netlink_oif_id)
                out_interface_name = link[0].get_attr("IFLA_IFNAME")
        except Exception:
            logger.exception(
                f"can't get output interface name of netlink interface with id {netlink_oif_id}"
            )
    return out_interface_name


def _signal_based_wait():
    exit_later = True

    def exit_gracefully(*args):
        nonlocal exit_later
        exit_later = False

    signal.signal(signal.SIGINT, exit_gracefully)
    signal.signal(signal.SIGTERM, exit_gracefully)

    while exit_later:
        time.sleep(1)


if (
    __name__ == "__main__"
):  # this module is library, run as script for development purposes
    graceful_shutdown_callback = start_route_reflector_in_background()
    _signal_based_wait()
    graceful_shutdown_callback()



"""
Provides utility functionality related to Pyro5 framework used for communication between client and server.
"""

import time
from abc import ABC

import Pyro5

from vrouter.common import serializers
from vrouter.common.utils import get_main_logger

logger = get_main_logger()
api_expose = Pyro5.api.expose


class NamedWrapper(ABC):
    @staticmethod
    def name():
        pass


def _init(wrappers, pyro_daemon_port, nameserver_daemon_port):
    """Configure and start Pyro5 instance (multiple daemons)"""

    # configure Pyro settings
    Pyro5.config.SERVERTYPE = "multiplex"
    Pyro5.config.POLLTIMEOUT = 3

    # fix serialization for custom types
    serializers.import_pyro5_serializers()

    # preparing host values for Namespace and Pyro5 main daemon
    # (Note: development and test/production environment is supported:
    # 1. development(Pyro5 runs locally): client connects to namespace daemon and due to its host set to
    # all_ip_addresses, connection from client is accepted. Nameserver provides address to exact object (exposed API)
    # in Pyro5 main daemon (object identified and pyro5 main daemon ip address and port). Client connect directly
    # to pyro5 main daemon and connection is accepted due to correct destination ip address.
    # 2. test/production (Pyro run in container/remotely): client connects to ip address/port of container
    # with namespace daemon (daemon accepts it due to all_ip_addresses set in as its host) and provides back address
    # to exact object (exposed API)in Pyro5 main daemon (object identified and pyro5 main daemon ip address and port).
    # It will reference containers IP address (+ port that should be open on container). Client will connect to
    # container ip and Pyro5 main daemon will see it coming from container interface with container IP address and
    # accept it.
    # Configuring Namespace daemon host to anything else (specific IP address) lead to problem in one of the cases.
    # )
    all_ip_addresses = "0.0.0.0"
    my_ip_address = Pyro5.socketutil.get_ip_address(None, workaround127=True)

    # start pyro name server with broadcast server
    nameserver_uri, nameserver_daemon, broadcast_server = Pyro5.nameserver.start_ns(
        host=all_ip_addresses, port=nameserver_daemon_port
    )
    assert broadcast_server is not None, "expect a broadcast server to be created"
    logger.info("created pyro nameserver, uri=%s" % nameserver_uri)

    # create a pyro daemon
    pyro_daemon = Pyro5.api.Daemon(host=my_ip_address, port=pyro_daemon_port)

    # register wrapper with pyro daemons
    for wrapper in wrappers:
        wrapper_uri = pyro_daemon.register(wrapper)
        logger.info(f"created pyro server for {wrapper.name()} (uri={wrapper_uri})")
        nameserver_daemon.nameserver.register(wrapper.name(), wrapper_uri)

    return (nameserver_daemon, broadcast_server, pyro_daemon)


def _run_loop(nameserver_daemon, broadcast_server, pyro_daemon):
    """Run Pyro5 request loop"""

    # Because this server runs the different daemons using the "multiplex" server type,
    # we can use the built in support (since Pyro 4.44) to combine multiple daemon event loops.
    # We can then simply run the event loop of the 'master daemon'. It will dispatch correctly.
    pyro_daemon.combine(nameserver_daemon)
    pyro_daemon.combine(broadcast_server)

    def loopcondition():
        logger.debug(f"{time.asctime()} Waiting for requests...")
        return True

    pyro_daemon.requestLoop(loopcondition)


def _cleanup(nameserver_daemon, broadcast_server, pyro_daemon):
    """Clean up any Pyro5 setup"""
    # clean up
    nameserver_daemon.close()
    broadcast_server.close()
    pyro_daemon.close()
    pyro_daemon.shutdown()
    logger.info("pyro daemons cleaned up")
    # kill pyro daemon


def get_external_ip_address():
    return Pyro5.socketutil.get_ip_address(None, workaround127=True)


def run_server_loop(*wrappers, pyro_daemon_port=8888, nameserver_daemon_port=9999):
    """Runs Pyro5 server and waits in loop for requests (blocking call)"""

    daemons = _init(wrappers, pyro_daemon_port, nameserver_daemon_port)
    _run_loop(*daemons)
    _cleanup(*daemons)  # for cases when loop breaks/is shutdown by call



"""
Handlers provide configuration-item-related functionality. See models.ConfigurationItemHandler .
"""

import base64
import ipaddress
import json
import os
import re
import socket
import subprocess
import time
import traceback
from datetime import datetime
from ipaddress import ip_address
from typing import List

from Pyro5.client import Proxy
from vpp_papi import VppEnum

from vrouter.common import serializers
from vrouter.common.models import (
    ConfigurationItemHandler,
    InterfaceConfigurationItem,
    ApplyReply,
    InterfaceType,
    SW_IF_INDEX,
    SWInterfaceDetail,
    RouteConfigurationItem,
    NO_INTERFACE_SW_INDEX,
    FibPathProto,
    FibPathType,
    FibPathFlags,
    IPRouteDetail,
    RouteType,
    ACLConfigurationItem,
    ACL_INDEX,
    ACLRuleConfigurationItem,
    ACLAction,
    OtherProtocol,
    ICMPProtocol,
    TCPProtocol,
    UDPProtocol,
    NO_ACL_INDEX,
    ACLDetails,
    DNat44ConfigurationItem,
    Nat44InterfaceConfigurationItem,
    Nat44AddressPoolConfigurationItem,
    NAT44AddressDetails,
    NAT44InterfaceDetails,
    NAT44StaticMappingDetails,
    NAT44IdentityMappingDetails,
    GREInterfaceLink,
    GRELinkType,
    WireguardInterfaceLink,
    WireguardPeerConfigurationItem,
    WIREGUARD_PEER_INDEX,
    NO_WIREGUARD_PEER_INDEX,
    WireguardPeerDetail,
    FRRConfigurationItem,
    LCPGlobalsConfigurationItem,
    LCPPairConfigurationItem,
    HOST_SW_IF_INDEX,
    LCPHostInterfaceTypeEnum,
    LCPPairDetail,
    VPPConfigUpdater,
    NOT_SET_MTU,
    RunningPacketCapture,
    PCAP_ANY_INTERFACE,
    L2PacketMatchingRule,
    L3PacketMatchingRule,
    NO_PROTOCOL,
    LinuxDockerVethTunnelConfigurationItem,
    LinuxRouteConfigurationItem,
    DHCPConfigurationItem,
    KeaConfigRequest,
    KeaConfigResponse,
)
from vrouter.common.utils import get_main_logger

logger = get_main_logger()
enabled_nat44_ed = False


def enforce_enabled_nat_in_vpp(function):
    """Decorator for enforcing NAT (NAT44ed) to be enabled in VPP"""

    def wrapper(*args, **kwds):
        self = args[0]
        global enabled_nat44_ed
        if not enabled_nat44_ed:
            reply = self.vpp_client.api.nat44_ed_plugin_enable_disable(enable=1)
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error="Failed to enable endpoint dependant nat44."
                    + get_vpp_error_log_message(reply.retval),
                )
            enabled_nat44_ed = True
        return function(*args, **kwds)

    return wrapper


def validate_config_item(function):
    """Decorator for checking configuration item argument to be handled by handler that is called"""

    def wrapper(*args, **kwds):
        self = args[0]
        configuration_item = args[1]
        if not self.is_handling_configuration_item(configuration_item):
            raise ValueError(
                f"configuration item is not handled by this handler ({self.__class__.__name__})"
            )
        return function(*args, **kwds)

    return wrapper


def validate_multiple_config_items(function):
    """Decorator for checking multiple configuration item arguments to be handled by handler that is called"""

    def wrapper(*args, **kwds):
        self = args[0]
        configuration_item = args[1]
        if not self.is_handling_configuration_item(configuration_item):
            raise ValueError(
                f"first configuration item is not handled by this handler ({self.__class__.__name__})"
            )
        configuration_item = args[2]
        if not self.is_handling_configuration_item(configuration_item):
            raise ValueError(
                f"second configuration item is not handled by this handler ({self.__class__.__name__})"
            )
        return function(*args, **kwds)

    return wrapper


vpp_error_codes_to_messages = {
    -1: "Unspecified Error",
    -2: "Invalid sw_if_index",
    -3: "No such FIB  VRF ",
    -4: " No such inner FIB  VRF  ",
    -5: " No such label",
    -6: "No such entry",
    -7: "Invalid value",
    -8: "Invalid value 2",
    -9: "Unimplemented",
    -10: "Invalid sw_if_index 2",
    -11: "System call error 1",
    -12: " System call error 2",
    -13: "System call error 3",
    -14: "System call error 4",
    -15: "System call error 5",
    -16: "System call error 6",
    -17: "System call error 7",
    -18: "System call error 8",
    -19: "System call error 9",
    -20: "System call error 10",
    -30: "Feature disabled by configuration",
    -31: "Invalid registration",
    -50: "Next hop not in FIB",
    -51: "Unknown destination",
    -52: "No paths specified in route",
    -53: " Next hop not found multipath",
    -54: "No matching interface for probe",
    -55: " Invalid VLAN",
    -56: "VLAN subif already exists",
    -57: "Invalid src address",
    -58: "Invalid dst address",
    -59: "Address length mismatch ",
    -60: "Address not found for interface",
    -61: "Address not deletable  ",
    -62: "ip6 not enabled",
    -63: "No such graph node",
    -64: "No such graph node 2",
    -65: " No such table",
    -66: "No such table 2",
    -67: "No such table 3",
    -68: "Subinterface already exists",
    -69: "Subinterface creation failed ",
    -70: "Invalid memory size requested",
    -71: "Invalid interface",
    -72: "Invalid number of tags for requested operation",
    -73: " Invalid argument",
    -74: "Unexpected interface state ",
    -75: " Tunnel already exists  ",
    -76: "Invalid decapnext",
    -77: "Response not ready ",
    -78: " Not connected to the data plane ",
    -79: " Interface already exists    ",
    -80: "Operation not allowed on slave of BondEthernet",
    -81: "Value already exists",
    -82: "Source and destination are the same",
    -83: "IP6 multicast address required",
    -84: "Segment routing policy name required",
    -85: "Not running as root",
    -86: "Connection to the data plane already exists",
    -87: "Unsupported JNI version",
    -88: "IP prefix invalid masked bits set in address",
    -89: " Invalid worker thread",
    -90: "LISP is disabled",
    -91: "Classify table not found",
    -92: " Unsupported LISP EID type",
    -93: "Cannot create pcap file ",
    -94: "Invalid adjacency type for this operation",
    -95: "Operation would exceed configured capacity of ranges",
    -96: "Operation would exceed capacity of number of ports",
    -97: "Invalid address family ",
    -98: "Invalid subinterface sw_if_index",
    -99: "Table too big",
    -100: "Cannot enabledisable feature ",
    -101: "Duplicate BFD object",
    -102: "No such BFD object ",
    -103: "BFD object in use ",
    -104: "BFD feature not supported",
    -105: "Address in use",
    -106: "Address not in use ",
    -107: "Queue full ",
    -108: "Unsupported application config",
    -109: "URI FIFO segment create failed ",
    -110: "RLOC address is local",
    -111: "BFD object cannot be manipulated at this time",
    -112: "Invalid GPE mode ",
    -113: "LISP GPE entries are present",
    -114: "Address found for interface ",
    -115: "Session failed to connect",
    -116: "Entry already exists  ",
    -117: "Svm segment create fail",
    -118: "Application not attached",
    -119: "Bridge domain already exists",
    -120: "Bridge domain has member interfaces  ",
    -121: "Bridge domain 0 cant be deletedmodified",
    -122: "Bridge domain ID exceeds 16M limit",
    -123: " Subinterface doesnt exist ",
    -124: "Client already exist for L2 MACs events",
    -125: "Invalid queue ",
    -126: "Unsupported",
    -127: "Address already present on another interface",
    -128: " Invalid application namespace",
    -129: "Wrong app namespace secret ",
    -130: "Connect scope  ",
    -131: " App already attached",
    -132: "Redirect failed ",
    -133: "Illegal name ",
    -134: "No name servers configured    ",
    -135: "Name server not found  ",
    -136: "Name resolution not enabled ",
    -137: "Server format error bug ",
    -138: "No such name",
    -139: "No addresses available",
    -140: "Retry with new server",
    -141: "Connect was filtered",
    -142: "Inbound ACL in use ",
    -143: "Outbound ACL in use ",
    -144: "Initialization Failed ",
    -145: " Netlink error",
    -146: "BIER bitstringlength unsupported",
    -147: "Instance in use",
    -148: "Session ID out of range",
    -149: " ACL in use by a lookup context",
    -150: " Invalid value 3",
    -151: "Interface is not an Ethernet interface   ",
    -152: "Bridge domain already has a BVI interface ",
    -153: "Invalid Protocol",
    -154: "Invalid Algorithm",
    -155: "Resource In Use ",
    -156: "invalid Key Length ",
    -157: "Unsupported FIB Path protocol",
    -159: "Endian mismatch detected ",
    -160: "No change in table",
    -161: "Missing certifcate or key  ",
    -162: "limit exceeded ",
    -163: "port not managed by IKE",
    -164: "UDP port already taken",
    -165: "Retry stream call with cursor",
    -166: "Invalid value 4 ",
    -167: "Busy",
    -168: "Bug",
}


def get_vpp_error_message(error_code: int) -> str:
    message = vpp_error_codes_to_messages.get(error_code)
    if message is None:
        return ""
    return message


def get_vpp_error_log_message(error_code: int) -> str:
    return f"VPP return error message: {get_vpp_error_message(error_code)} (return code {error_code})."


class InterfaceHandler(VPPConfigUpdater, ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, InterfaceConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return {f"interface/{configuration_item.name}"}

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        match configuration_item.type:
            case InterfaceType.TAP:
                reply = self.vpp_client.api.tap_create_v2()
            case InterfaceType.GRE_TUNNEL:
                gre_type_config_to_vpp = {
                    GRELinkType.L3: 0,
                    GRELinkType.TEB: 1,
                    GRELinkType.ERSPAN: 2,
                }
                if not isinstance(configuration_item.link, GREInterfaceLink):
                    return ApplyReply(
                        success=False,
                        error="Failed to create a GRE interface, because no "
                        "GRE-specific information (InterfaceConfigurationItem.link) "
                        "was present",
                    )
                reply = self.vpp_client.api.gre_tunnel_add_del(
                    is_add=1,
                    tunnel={
                        "src": configuration_item.link.src_addr,
                        "dst": configuration_item.link.dst_addr,
                        "outer_table_id": configuration_item.link.outer_fib_id,
                        "instance": 0xFFFFFFFF,
                        "type": gre_type_config_to_vpp[configuration_item.link.type],
                        "session_id": configuration_item.link.session_id,
                        # "mode": ...
                        # "flags": ...
                    },
                )
                if reply.retval == 0:
                    reply_from_tagging = self._add_tag_to_interface(
                        tag=configuration_item.name,
                        sw_if_index=reply.sw_if_index,
                        is_add=True,
                    )
                    if reply_from_tagging.retval != 0:
                        return ApplyReply(
                            success=False,
                            error="Failed to tag GRE interface."
                            + get_vpp_error_log_message(reply_from_tagging.retval),
                        )
            case InterfaceType.WIREGUARD_TUNNEL:
                if not isinstance(configuration_item.link, WireguardInterfaceLink):
                    return ApplyReply(
                        success=False,
                        error="Failed to create wireguard interface, because "
                        "no wireguard-specific information ("
                        "InterfaceConfigurationItem.link) was present",
                    )
                reply = self.vpp_client.api.wireguard_interface_create(
                    interface={
                        "user_instance": 0xFFFFFFFF,
                        "port": configuration_item.link.port,
                        "src_ip": configuration_item.link.src_addr,
                        "private_key": base64.b64decode(
                            configuration_item.link.private_key
                        ),
                        # 'public_key': ...
                        # 'sw_if_index': ...
                    },
                    generate_key=False,
                )
                if reply.retval == 0:
                    reply_from_tagging = self._add_tag_to_interface(
                        tag=configuration_item.name,
                        sw_if_index=reply.sw_if_index,
                        is_add=True,
                    )
                    if reply_from_tagging.retval != 0:
                        return ApplyReply(
                            success=False,
                            error="Failed to tag wireguard interface."
                            + get_vpp_error_log_message(reply_from_tagging.retval),
                        )
            case InterfaceType.SOFTWARE_LOOPBACK:
                reply = self.vpp_client.api.create_loopback()
            case _:
                return ApplyReply(
                    success=False,
                    error="Interface type {} is not supported".format(
                        configuration_item.type
                    ),
                )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create interface."
                + get_vpp_error_log_message(reply.retval),
            )

        if_index = reply.sw_if_index
        if configuration_item.enabled:
            reply = self.vpp_client.api.sw_interface_set_flags(
                sw_if_index=if_index,
                flags=1,  # for dump use VppEnum.vl_api_if_status_flags_t.IF_STATUS_API_FLAG_LINK_UP
            )
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error=f"Failed to enable interface(sw_if_index {if_index}). "
                    + get_vpp_error_log_message(reply.retval),
                )

        if configuration_item.ip_addresses:
            for ip_address in configuration_item.ip_addresses:
                reply = self.vpp_client.api.sw_interface_add_del_address(
                    sw_if_index=if_index,
                    is_add=True,
                    prefix=ip_address,  # i.e. "10.10.10.10/16"
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to add ip address {ip_address} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )

        if configuration_item.mtu != NOT_SET_MTU:
            # Note: there are multiple MTU settings, see https://docs.fd.io/vpp/19.01/md_src_vnet_MTU.html
            # Ligato's VPP-agent sets only HW MTU, but we need to setup also Wireguard MTU and that interface is
            # virtual, so only SW MTU can be configured
            if not self._supports_setting_hw_mtu(configuration_item.type):
                # setting of HW MTU is unsupported -> setting up SW MTU
                reply = self.vpp_client.api.sw_interface_set_mtu(
                    sw_if_index=if_index,
                    mtu=[
                        configuration_item.mtu,  # L3 MTU
                        configuration_item.mtu,  # IPv4 MTU
                        configuration_item.mtu,  # IPv6 MTU
                        configuration_item.mtu,  # MPLS MTU
                    ],
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to set SW MTU {configuration_item.mtu} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )
            else:
                # Note: setting up VPP's HW MTU -> all SW MTUs will be set to HW MTU
                reply = self.vpp_client.api.hw_interface_set_mtu(
                    sw_if_index=if_index, mtu=configuration_item.mtu
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to set HW MTU {configuration_item.mtu} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )

        return ApplyReply(success=True, vpp_data={SW_IF_INDEX: if_index})

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        sw_if_index = self._get_sw_if_index(configuration_item.name)
        if sw_if_index == NO_INTERFACE_SW_INDEX:
            return ApplyReply(
                success=False,
                error=f"Can't find index for interface {configuration_item.name}. ",
            )

        match configuration_item.type:
            case InterfaceType.TAP:
                reply = self.vpp_client.api.tap_delete_v2(sw_if_index=sw_if_index)
            case InterfaceType.GRE_TUNNEL:
                gre_type_config_to_vpp = {
                    GRELinkType.L3: 0,
                    GRELinkType.TEB: 1,
                    GRELinkType.ERSPAN: 2,
                }
                reply = self.vpp_client.api.gre_tunnel_add_del(
                    is_add=0,
                    tunnel={
                        "src": configuration_item.link.src_addr,
                        "dst": configuration_item.link.dst_addr,
                        "outer_table_id": configuration_item.link.outer_fib_id,
                        "instance": 0xFFFFFFFF,
                        "type": gre_type_config_to_vpp[configuration_item.link.type],
                        "session_id": configuration_item.link.session_id,
                        # "mode": ...
                        # "flags": ...
                    },
                )
                if reply.retval == 0:
                    reply_from_untagging = self._add_tag_to_interface(
                        tag=configuration_item.name, is_add=False
                    )
                    if reply_from_untagging.retval != 0:
                        return ApplyReply(
                            success=False,
                            error="Failed to untag GRE interface."
                            + get_vpp_error_log_message(reply_from_untagging.retval),
                        )
            case InterfaceType.WIREGUARD_TUNNEL:
                reply = self.vpp_client.api.wireguard_interface_delete(
                    sw_if_index=sw_if_index
                )
                if reply.retval == 0:
                    reply_from_untagging = self._add_tag_to_interface(
                        tag=configuration_item.name, is_add=False
                    )
                    if reply_from_untagging.retval != 0:
                        return ApplyReply(
                            success=False,
                            error="Failed to untag wireguard interface."
                            + get_vpp_error_log_message(reply_from_untagging.retval),
                        )
            case InterfaceType.SOFTWARE_LOOPBACK:
                reply = self.vpp_client.api.delete_loopback(sw_if_index=sw_if_index)
            case _:
                return ApplyReply(
                    success=False,
                    error="Removal of interface type {} is not supported".format(
                        configuration_item.type
                    ),
                )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error=f"Failed to remove interface {configuration_item.name}."
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)

    @validate_multiple_config_items
    def update_in_vpp(self, old_configuration_item, new_configuration_item):
        # test whether only supported changes are made
        old = old_configuration_item.to_dict()
        old.update(
            {"enabled": True, "ip_addresses": [], "mtu": NOT_SET_MTU}
        )  # fill updatable fields with equal values
        new = new_configuration_item.to_dict()
        new.update(
            {"enabled": True, "ip_addresses": [], "mtu": NOT_SET_MTU}
        )  # fill updatable fields with equal values
        if old != new:
            return ApplyReply(
                success=False,
                error="Only update in fields 'enabled', 'ip_addresses', 'mtu' are supported",
            )

        # set status (if changed)
        if_index = self._get_sw_if_index(new_configuration_item.name)
        if old_configuration_item.enabled != new_configuration_item.enabled:
            reply = self.vpp_client.api.sw_interface_set_flags(
                sw_if_index=if_index, flags=1 if new_configuration_item.enabled else 0
            )
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error=f"Failed to enable/disable interface(sw_if_index {if_index}). "
                    + get_vpp_error_log_message(reply.retval),
                )
        # set ip addresses (if changed)
        for addr in (
            old_configuration_item.ip_addresses + new_configuration_item.ip_addresses
        ):
            if (
                addr in new_configuration_item.ip_addresses
                and addr not in old_configuration_item.ip_addresses
            ):
                reply = self.vpp_client.api.sw_interface_add_del_address(
                    sw_if_index=if_index,
                    is_add=True,
                    prefix=addr,  # i.e. "10.10.10.10/16"
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to add ip address {addr} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )
            elif (
                addr not in new_configuration_item.ip_addresses
                and addr in old_configuration_item.ip_addresses
            ):
                reply = self.vpp_client.api.sw_interface_add_del_address(
                    sw_if_index=if_index,
                    is_add=False,
                    prefix=addr,  # i.e. "10.10.10.10/16"
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to remove ip address {addr} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )
        if (
            old_configuration_item.mtu != new_configuration_item.mtu
            and new_configuration_item.mtu != NOT_SET_MTU
        ):
            if not self._supports_setting_hw_mtu(new_configuration_item.type):
                # setting of HW MTU is unsupported -> setting up SW MTU
                reply = self.vpp_client.api.sw_interface_set_mtu(
                    sw_if_index=if_index,
                    mtu=[
                        new_configuration_item.mtu,  # L3 MTU
                        new_configuration_item.mtu,  # IPv4 MTU
                        new_configuration_item.mtu,  # IPv6 MTU
                        new_configuration_item.mtu,  # MPLS MTU
                    ],
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to update SW MTU {new_configuration_item.mtu} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )
            else:
                # Note: setting up VPP's HW MTU -> all SW MTUs will be set to HW MTU
                reply = self.vpp_client.api.hw_interface_set_mtu(
                    sw_if_index=if_index, mtu=new_configuration_item.mtu
                )
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to update HW MTU {new_configuration_item.mtu} to "
                        + f"interface(sw_if_index {if_index}). "
                        + get_vpp_error_log_message(reply.retval),
                    )

        return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        sw_if_index = self._get_sw_if_index(configuration_item.name)
        for interface_detail in self.vpp_client.api.sw_interface_dump():
            if interface_detail.sw_if_index == sw_if_index:
                return SWInterfaceDetail.get_from_basic_vpp_client_equivalent_obj(
                    interface_detail
                )
        return dict()

    def _add_tag_to_interface(
        self, tag, sw_if_index=NO_INTERFACE_SW_INDEX, is_add=True
    ):
        return self.vpp_client.api.sw_interface_tag_add_del(
            tag=tag,
            is_add=1 if is_add else 0,
            # For some reason, if deleting tag, the software interface index has to be 0 and only name should be set.
            # Otherwise reply returns with error core -2 (incorrect sw_if_idx)
            sw_if_index=sw_if_index if is_add else 0,
        )

    def _supports_setting_hw_mtu(self, interface_type):
        # Note: virtual interfaces have no HW, hence can't set HW MTU
        return interface_type not in (
            InterfaceType.UNDEFINED_TYPE,
            InterfaceType.VXLAN_TUNNEL,
            InterfaceType.IPSEC_TUNNEL,
            InterfaceType.GRE_TUNNEL,
            InterfaceType.WIREGUARD_TUNNEL,
            InterfaceType.SOFTWARE_LOOPBACK,
            InterfaceType.SUB_INTERFACE,
        )


class RouteHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, RouteConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone, but it is normally something like this:
        # {f"route/vrf/<vrf_id>/if/<out_if>/dest/<dest_network>/gw/<next_hop>"}

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        if configuration_item.multi_output_paths:
            deps = set()
            for multipath in configuration_item.multi_output_paths:
                if multipath.outgoing_interface:
                    deps.add(f"interface/{multipath.outgoing_interface}")
            return deps
        if configuration_item.outgoing_interface:
            return {f"interface/{configuration_item.outgoing_interface}"}
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        return self._add_delete_route_in_vpp(configuration_item, is_add=True)

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        return self._add_delete_route_in_vpp(configuration_item, is_add=False)

    def _add_delete_route_in_vpp(self, configuration_item, is_add):
        path_type = (
            FibPathType.FIB_API_PATH_TYPE_DROP
            if configuration_item.type == RouteType.DROP
            else FibPathType.FIB_PATH_TYPE_NORMAL
        )
        sw_if_index = 0

        route = {
            "table_id": 0,  # hardcoded as there are not used multiple VRFs
            "prefix": configuration_item.destination_network,
            "n_paths": 0,
            "paths": [],
        }
        if not configuration_item.multi_output_paths:  # single path
            if configuration_item.type == RouteType.INTRA_VRF:
                sw_if_index = NO_INTERFACE_SW_INDEX
                if configuration_item.outgoing_interface:
                    sw_if_index = self._get_sw_if_index(
                        configuration_item.outgoing_interface
                    )
            path = {
                "weight": configuration_item.weight,
                "preference": 0,
                "table_id": 0,
                "next_hop_id": 0xFFFFFFFF,  # INVALID_INDEX
                "sw_if_index": sw_if_index,
                "rpf_id": 0,
                "proto": FibPathProto.FIB_PATH_NH_PROTO_IP4,
                "type": path_type,
                "flags": FibPathFlags.FIB_PATH_FLAG_NONE,
                "n_labels": 0,
                "label_stack": [{}] * 16,
            }
            if configuration_item.next_hop_address:
                path["nh"] = {
                    "address": {
                        "ip4": ip_address(str(configuration_item.next_hop_address))
                    }
                }
            route["n_paths"] = 1
            route["paths"].append(path)
        else:  # multipath
            for config_path in configuration_item.multi_output_paths:
                if configuration_item.type == RouteType.INTRA_VRF:
                    sw_if_index = NO_INTERFACE_SW_INDEX
                    if config_path.outgoing_interface:
                        sw_if_index = self._get_sw_if_index(
                            config_path.outgoing_interface
                        )
                path = {
                    "weight": config_path.weight,
                    "preference": 0,
                    "table_id": 0,
                    "next_hop_id": 0xFFFFFFFF,  # INVALID_INDEX
                    "sw_if_index": sw_if_index,
                    "rpf_id": 0,
                    "proto": FibPathProto.FIB_PATH_NH_PROTO_IP4,
                    "type": path_type,
                    "flags": FibPathFlags.FIB_PATH_FLAG_NONE,
                    "n_labels": 0,
                    "label_stack": [{}] * 16,
                }
                if config_path.next_hop_address:
                    path["nh"] = {
                        "address": {
                            "ip4": ip_address(str(config_path.next_hop_address))
                        }
                    }
                route["paths"].append(path)
            route["n_paths"] = len(route["paths"])

        reply = self.vpp_client.api.ip_route_add_del(
            is_add=1 if is_add else 0,
            is_multipath=1 if configuration_item.multi_output_paths else 0,
            route=route,
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create route."
                + get_vpp_error_log_message(reply.retval),
            )

        return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        outgoing_interface_index = NO_INTERFACE_SW_INDEX
        if configuration_item.outgoing_interface:
            outgoing_interface_index = self._get_sw_if_index(
                configuration_item.outgoing_interface
            )

        dump = self.vpp_client.api.ip_route_dump()
        try:
            for route_detail in dump:
                if (
                    route_detail.route.table_id == 0
                    and route_detail.route.prefix
                    == ipaddress.IPv4Network(configuration_item.destination_network)
                ):
                    return IPRouteDetail.get_from_basic_vpp_client_equivalent_obj(
                        route_detail
                    )
        except Exception:
            logger.exception(f"route detail dumping failed for {configuration_item}")

        return dict()


class ACLHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index, get_acl_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index
        self._get_acl_index = get_acl_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, ACLConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        dependencies = set()
        for egress_interface_name in configuration_item.egress:
            dependencies.add(f"interface/{egress_interface_name}")
        for ingress_interface_name in configuration_item.ingress:
            dependencies.add(f"interface/{ingress_interface_name}")
        return dependencies

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        # create ACL in VPP
        rules = ACLHandler._transform_rules_to_vpp_rules(configuration_item.rules)
        if not rules:
            return ApplyReply(
                success=False, error=f"No rules found for ACL {configuration_item.name}"
            )
        reply = self.vpp_client.api.acl_add_replace(
            acl_index=0xFFFFFFFF,  # to make new Entry
            tag=configuration_item.name,
            count=len(rules),
            r=rules,
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create ACL." + get_vpp_error_log_message(reply.retval),
            )

        # link interfaces to ACL
        acl_index = reply.acl_index
        reply = self.link_unlink_acl_interfaces(
            configuration_item, acl_index, is_add=True
        )
        if not reply.success:
            return reply

        return ApplyReply(success=True, vpp_data={ACL_INDEX: acl_index})

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        acl_index = self._get_acl_index(configuration_item.name)
        if acl_index == NO_ACL_INDEX:
            return ApplyReply(
                success=False,
                error=f"Can't get index for ACL {configuration_item.name}",
            )

        # unlink interfaces linked to ACL
        reply = self.link_unlink_acl_interfaces(
            configuration_item, acl_index, is_add=False
        )
        if not reply.success:
            return reply

        # remove ACL from VPP
        reply = self.vpp_client.api.acl_del(acl_index=acl_index)
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to remove ACL {configuration_item.name}. "
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True, vpp_data={ACL_INDEX: acl_index})

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        acl_index = self._get_acl_index(configuration_item.name)
        if acl_index == NO_ACL_INDEX:
            return dict()

        reply = self.vpp_client.api.acl_dump(acl_index=acl_index)
        if len(reply) == 1:
            return ACLDetails.get_from_basic_vpp_client_equivalent_obj(reply[0])
        # TODO dumping ACL interfaces? we get only their sw_if_index and we have that remembered, so real value is only
        #  simple boolean check that it is there
        # self.vpp_client.api.acl_interface_list_dump(...)
        return dict()

    def link_unlink_acl_interfaces(self, configuration_item, acl_index, is_add=True):
        link_unlink = "link" if is_add else "unlink"
        for interface_name in configuration_item.egress:
            reply = self.vpp_client.api.acl_interface_add_del(
                is_add=is_add,
                acl_index=acl_index,
                sw_if_index=self._get_sw_if_index(interface_name),
                is_input=False,
            )
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error=f"Failed to {link_unlink} interface {interface_name} to ACL "
                    f"{configuration_item.name} as egress interface. "
                    + get_vpp_error_log_message(reply.retval),
                )
        for interface_name in configuration_item.ingress:
            reply = self.vpp_client.api.acl_interface_add_del(
                is_add=is_add,
                acl_index=acl_index,
                sw_if_index=self._get_sw_if_index(interface_name),
                is_input=True,
            )
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error=f"Failed to {link_unlink} interface {interface_name} to ACL "
                    f"{configuration_item.name} as ingress interface. "
                    + get_vpp_error_log_message(reply.retval),
                )
        return ApplyReply(success=True)

    @staticmethod
    def _transform_rules_to_vpp_rules(
        config_rules: List[ACLRuleConfigurationItem],
    ) -> List:
        config_action_to_vpp_action = {
            # Note: coincidentally ACLAction int value is the needed
            # VPP constant, but lets not interchange these 2 logically different things
            ACLAction.DENY: 0,
            ACLAction.PERMIT: 1,
            ACLAction.REFLECT: 2,
        }
        vpp_rules = []
        for config_rule in config_rules:
            vpp_rule = {"is_permit": config_action_to_vpp_action[config_rule.action]}
            if (
                config_rule.refinement is not None
            ):  # further refinement/narrowing of traffic that this rule is for
                if config_rule.refinement.addresses is not None:
                    vpp_rule["src_prefix"] = (
                        config_rule.refinement.addresses.source_network
                    )
                    vpp_rule["dst_prefix"] = (
                        config_rule.refinement.addresses.destination_network
                    )
                match config_rule.refinement.protocol:
                    case ICMPProtocol(
                        icmpv6=icmpv6,
                        icmp_code_range=icmp_code_range,
                        icmp_type_range=icmp_type_range,
                    ):
                        vpp_rule["proto"] = (
                            VppEnum.vl_api_ip_proto_t.IP_API_PROTO_ICMP6
                            if icmpv6
                            else VppEnum.vl_api_ip_proto_t.IP_API_PROTO_ICMP
                        )
                        vpp_rule["srcport_or_icmptype_first"] = icmp_type_range.first
                        vpp_rule["srcport_or_icmptype_last"] = icmp_type_range.last
                        vpp_rule["dstport_or_icmpcode_first"] = icmp_code_range.first
                        vpp_rule["dstport_or_icmpcode_last"] = icmp_code_range.last
                    case TCPProtocol(
                        destination_port_range=destination_port_range,
                        source_port_range=source_port_range,
                        tcp_flags_mask=tcp_flags_mask,
                        tcp_flags_value=tcp_flags_value,
                    ):
                        vpp_rule["proto"] = VppEnum.vl_api_ip_proto_t.IP_API_PROTO_TCP
                        vpp_rule["srcport_or_icmptype_first"] = (
                            source_port_range.lower_port
                        )
                        vpp_rule["srcport_or_icmptype_last"] = (
                            source_port_range.upper_port
                        )
                        vpp_rule["dstport_or_icmpcode_first"] = (
                            destination_port_range.lower_port
                        )
                        vpp_rule["dstport_or_icmpcode_last"] = (
                            destination_port_range.upper_port
                        )
                        vpp_rule["tcp_flags_mask"] = tcp_flags_mask
                        vpp_rule["tcp_flags_value"] = tcp_flags_value
                    case UDPProtocol(
                        destination_port_range=destination_port_range,
                        source_port_range=source_port_range,
                    ):
                        vpp_rule["proto"] = VppEnum.vl_api_ip_proto_t.IP_API_PROTO_UDP
                        vpp_rule["srcport_or_icmptype_first"] = (
                            source_port_range.lower_port
                        )
                        vpp_rule["srcport_or_icmptype_last"] = (
                            source_port_range.upper_port
                        )
                        vpp_rule["dstport_or_icmpcode_first"] = (
                            destination_port_range.lower_port
                        )
                        vpp_rule["dstport_or_icmpcode_last"] = (
                            destination_port_range.upper_port
                        )
                    case OtherProtocol(protocol=protocol):
                        vpp_rule["proto"] = protocol  # VppEnum.vl_api_ip_proto_t
            vpp_rules.append(vpp_rule)
        return vpp_rules


class Nat44AddressPoolHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client):
        self.vpp_client = vpp_client

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, Nat44AddressPoolConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        # ignoring vrf dependency as only one default vrf is used
        # ignoring need for NAT's endpoint-dependent mode as that is the default run of VPP
        return set()

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def add_to_vpp(self, configuration_item):
        return self.add_del_to_vpp(configuration_item, is_add=True)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def remove_from_vpp(self, configuration_item):
        return self.add_del_to_vpp(configuration_item, is_add=False)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def dump_from_vpp(self, configuration_item):
        return [
            NAT44AddressDetails.get_from_basic_vpp_client_equivalent_obj(address_detail)
            for address_detail in self.vpp_client.api.nat44_address_dump()
            if configuration_item.first_ip <= address_detail.ip_address
            and address_detail.ip_address <= configuration_item.last_ip
        ]

    def add_del_to_vpp(self, configuration_item, is_add=True):
        reply = self.vpp_client.api.nat44_add_del_address_range(
            is_add=1 if is_add else 0,
            first_ip_address=configuration_item.first_ip,
            last_ip_address=configuration_item.last_ip,
            vrf_id=0,  # we use only vrf 0 for failover case
            # not using twice-nat -> not configuring it
            # flags=VppEnum.vl_api_nat_config_flags_t.NAT_IS_TWICE_NAT if configuration_item.twice_nat else 0
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create NAT44 address pool."
                + get_vpp_error_log_message(reply.retval),
            )

        return ApplyReply(success=True)


class Nat44InterfaceHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, Nat44InterfaceConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return {f"interface/{configuration_item.name}"}

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def add_to_vpp(self, configuration_item):
        return self.add_del_to_vpp(configuration_item, is_add=True)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def remove_from_vpp(self, configuration_item):
        return self.add_del_to_vpp(configuration_item, is_add=False)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def dump_from_vpp(self, configuration_item):
        interface_index = self._get_sw_if_index(configuration_item.name)
        for detail in self.vpp_client.api.nat44_interface_dump():
            if detail.sw_if_index == interface_index:
                return NAT44InterfaceDetails.get_from_basic_vpp_client_equivalent_obj(
                    detail
                )

    def add_del_to_vpp(self, configuration_item, is_add=True):
        if configuration_item.nat_inside:
            reply = self.enable_nat44_interface(
                configuration_item.name,
                True,
                configuration_item.output_feature,
                is_add=is_add,
            )
            if not reply.success:
                return reply
        if configuration_item.nat_outside:
            reply = self.enable_nat44_interface(
                configuration_item.name,
                False,
                configuration_item.output_feature,
                is_add=is_add,
            )
            if not reply.success:
                return reply
        if (
            configuration_item.output_feature
            and not configuration_item.nat_inside
            and not configuration_item.nat_outside
        ):
            reply = self.enable_nat44_interface(
                configuration_item.name,
                False,
                configuration_item.output_feature,
                is_add=is_add,
            )
            if not reply.success:
                return reply
        return ApplyReply(success=True)

    def enable_nat44_interface(self, name, is_inside, is_output_feature, is_add=True):
        create_delete = "create" if is_add else "delete"
        add_del_feature = self.vpp_client.api.nat44_interface_add_del_feature
        if is_output_feature:
            add_del_feature = self.vpp_client.api.nat44_interface_add_del_output_feature
        reply = add_del_feature(
            sw_if_index=self._get_sw_if_index(name),
            flags=VppEnum.vl_api_nat_config_flags_t.NAT_IS_INSIDE if is_inside else 0,
            is_add=is_add,
        )
        if reply.retval != 0:
            output_feature = " (output feature)" if is_output_feature else ""
            return ApplyReply(
                success=False,
                error="Failed to {create_delete} nat interface{output_feature}. "
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)


class DNat44Handler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, DNat44ConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        # ignoring need for NAT's endpoint-dependent mode as that is the default run of VPP
        # ignoring nat pool dependency for twice-nat as not used in failover use case
        # ignoring vrf dependency as only one vrf is used and that is the default one that is always present
        needed_interfaces = {
            sm.external_interface
            for sm in configuration_item.static_mappings
            if sm.external_interface != ""
        }.union(
            {
                im.interface
                for im in configuration_item.identity_mappings
                if im.interface != ""
            }
        )
        return {
            f"interface/{needed_interface_name}"
            for needed_interface_name in needed_interfaces
        }

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def add_to_vpp(self, configuration_item):
        return self.add_delete_from_vpp(configuration_item, is_add=True)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def remove_from_vpp(self, configuration_item):
        return self.add_delete_from_vpp(configuration_item, is_add=False)

    @validate_config_item
    @enforce_enabled_nat_in_vpp
    def dump_from_vpp(self, configuration_item):
        return {
            "static_mappings": [
                NAT44StaticMappingDetails.get_from_basic_vpp_client_equivalent_obj(
                    static_mapping
                )
                for static_mapping in self.vpp_client.api.nat44_static_mapping_dump()
                if static_mapping.tag == configuration_item.label
            ],
            "identity_mappings": [
                NAT44IdentityMappingDetails.get_from_basic_vpp_client_equivalent_obj(
                    identity_mapping
                )
                for identity_mapping in self.vpp_client.api.nat44_identity_mapping_dump()
                if identity_mapping.tag == configuration_item.label
            ],
        }

    def add_delete_from_vpp(self, configuration_item, is_add=True):
        for identity_mapping in configuration_item.identity_mappings:
            reply = self.add_del_identity_mapping(
                identity_mapping, configuration_item.label, is_add=is_add
            )
            if not reply.success:
                return reply
        for static_mapping in configuration_item.static_mappings:
            reply = self.add_del_static_mapping(
                static_mapping, configuration_item.label, is_add=is_add
            )
            if not reply.success:
                return reply
        return ApplyReply(success=True)

    def add_del_identity_mapping(self, identity_mapping, nat_label, is_add=True):
        interface_sw_index = NO_INTERFACE_SW_INDEX
        if identity_mapping.interface:
            interface_sw_index = self._get_sw_if_index(identity_mapping.interface)
            if (
                interface_sw_index == NO_INTERFACE_SW_INDEX
            ):  # not found index for interface name
                return ApplyReply(
                    success=False,
                    error=f"Can't find sw_if_index for interface with name {identity_mapping.interface}",
                )
        reply = self.vpp_client.api.nat44_add_del_identity_mapping(
            is_add=1 if is_add else 0,
            tag=nat_label,
            flags=(
                VppEnum.vl_api_nat_config_flags_t.NAT_IS_ADDR_ONLY
                if identity_mapping.port == 0
                else 0
            ),
            ip_address=(
                identity_mapping.ip_address
                if not identity_mapping.interface
                else ipaddress.IPv4Address("0.0.0.0")
            ),
            port=identity_mapping.port,
            protocol=int(identity_mapping.protocol),
            sw_if_index=interface_sw_index,
            vrf_id=0,  # using always vrf 0 for failover use case
        )
        if reply.retval != 0:
            add_remove = "add" if is_add else "remove"
            return ApplyReply(
                success=False,
                error=f"Failed to {add_remove} identity mapping ({identity_mapping}). "
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)

    def add_del_static_mapping(self, static_mapping, nat_label, is_add=True):
        if not static_mapping.local_ips:
            return ApplyReply(success=False, error="No local address provided")
        if len(static_mapping.local_ips) > 1:
            # Note: we should not need it in failover use case
            return ApplyReply(success=False, error="Loadbalanced NAT is not supported")

        external_interface_sw_index = 0  # NOTE: This is a workaround, because NO_INTERFACE_SW_INDEX crashes VPP 22.02 # TODO still problem with currently used VPP 22.10.1 ?
        if static_mapping.external_interface:
            external_interface_sw_index = self._get_sw_if_index(
                static_mapping.external_interface
            )
            if (
                external_interface_sw_index == NO_INTERFACE_SW_INDEX
            ):  # not found index for interface name
                return ApplyReply(
                    success=False,
                    error=f"Can't find sw_if_index for interface with name "
                    f"{static_mapping.external_interface}",
                )

        # ignoring twicenat flags as twice nat is not needed for failover case
        # isTwiceNat:     mapping.TwiceNat == nat.DNat44_StaticMapping_ENABLED,
        # isSelfTwiceNat: mapping.TwiceNat == nat.DNat44_StaticMapping_SELF,
        flags = VppEnum.vl_api_nat_config_flags_t.NAT_IS_OUT2IN_ONLY
        if (
            static_mapping.local_ips[0].local_port == 0
            and static_mapping.external_port == 0
        ):
            flags |= VppEnum.vl_api_nat_config_flags_t.NAT_IS_ADDR_ONLY

        reply = self.vpp_client.api.nat44_add_del_static_mapping_v2(
            is_add=is_add,
            tag=nat_label,
            local_ip_address=static_mapping.local_ips[0].local_ip,
            external_ip_address=(
                static_mapping.external_ip
                if not static_mapping.external_interface
                else ipaddress.IPv4Address("0.0.0.0")
            ),
            protocol=int(static_mapping.protocol),
            external_sw_if_index=external_interface_sw_index,
            vrf_id=0,  # only VRF 0 is used in failover use case
            flags=flags,
            local_port=static_mapping.local_ips[0].local_port,
            external_port=static_mapping.external_port,
            # Note: not setting twiceNat pool ip as twice nat is not needed for failover case
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error=f"Failed to create static mapping ({static_mapping}). "
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)


class WireguardPeerHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index, get_wireguard_peer_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index
        self._get_wireguard_peer_index = get_wireguard_peer_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, WireguardPeerConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        if configuration_item.wg_if_name:
            return {f"interface/{configuration_item.wg_if_name}"}
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        reply = self.vpp_client.api.wireguard_peer_add(
            peer={
                "public_key": base64.b64decode(configuration_item.public_key),
                "port": configuration_item.port,
                "endpoint": configuration_item.endpoint,
                "n_allowed_ips": len(configuration_item.allowed_ips),
                "allowed_ips": configuration_item.allowed_ips,
                "sw_if_index": self._get_sw_if_index(configuration_item.wg_if_name),
                "persistent_keepalive": configuration_item.persistent_keepalive,
                # ignoring: flags, table_id, peer_index
            }
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create wireguard peer."
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(
            success=True, vpp_data={WIREGUARD_PEER_INDEX: reply.peer_index}
        )

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        peer_index = self._get_wireguard_peer_index(configuration_item)
        if peer_index == NO_WIREGUARD_PEER_INDEX:
            ApplyReply(success=False, error="Peer index not found")
        reply = self.vpp_client.api.wireguard_peer_remove(peer_index=peer_index)
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error=f"Failed to remove wireguard peer (peer_index {peer_index}). "
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        peer_index = self._get_wireguard_peer_index(configuration_item)
        if peer_index == NO_WIREGUARD_PEER_INDEX:
            logger.warning(
                f"can't find index for wireguard peer {configuration_item.name} and "
                f"therefore no vpp dump will be available"
            )
            return dict()
        reply = self.vpp_client.api.wireguard_peers_dump(peer_index=peer_index)
        if not reply:
            logger.warning(
                f"vpp can't find data for given wireguard peer index {peer_index} and "
                f"therefore no vpp dump will be available"
            )
            return dict()
        return WireguardPeerDetail.get_from_basic_vpp_client_equivalent_obj(
            reply[0].peer
        )


class LCPHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(
            configuration_item, LCPGlobalsConfigurationItem
        ) or isinstance(configuration_item, LCPPairConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # noone uses it

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        if isinstance(configuration_item, LCPPairConfigurationItem):
            return {f"interface/{configuration_item.interface}"}
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        if isinstance(configuration_item, LCPPairConfigurationItem):
            return self.add_del_lcp_pair_to_vpp(configuration_item, is_add=True)

        # add LCP Global settings
        reply = self.vpp_client.api.cli_inband(
            cmd=f"lcp default netns {configuration_item.default_namespace}"
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to set default namespace for LCP."
                + get_vpp_error_log_message(reply.retval),
            )
        reply = self.vpp_client.api.cli_inband(
            cmd="lcp lcp-sync {}".format("on" if configuration_item.lcp_sync else "off")
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to set lcp-sync."
                + get_vpp_error_log_message(reply.retval),
            )
        reply = self.vpp_client.api.cli_inband(
            cmd="lcp lcp-auto-subint {}".format(
                "on" if configuration_item.lcp_auto_subint else "off"
            )
        )
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to set lcp-auto-subint."
                + get_vpp_error_log_message(reply.retval),
            )
        return ApplyReply(success=True)

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        if isinstance(configuration_item, LCPPairConfigurationItem):
            return self.add_del_lcp_pair_to_vpp(configuration_item, is_add=False)
        # there is no remove for LCPGlobals (they can be only set and we don't "unset" them for delete)
        return ApplyReply(success=True)

    def add_del_lcp_pair_to_vpp(self, configuration_item, is_add=True):
        # the binary api for linux-cp's pair add/del is broken (https://lists.fd.io/g/vpp-dev/topic/lcp_plugin_api_issue_report/89972222?p=)
        # reply = self.vpp_client.api.lcp_itf_pair_add_del_v2(
        #     is_add=1 if is_add else 0,
        #     sw_if_index=self._get_sw_if_index(configuration_item.interface),
        #     host_if_name=configuration_item.mirror_interface_host_name,
        #     host_if_type=int(configuration_item.mirror_interface_type),
        #     netns=configuration_item.host_namespace
        # )
        # using CLI instead (CLI )
        #   lcp create <sw_if_index>|<if-name> host-if <host-if-name> netns <namespace> [tun]
        #    <sw_if_index>|<if-name>
        if is_add:
            cmd = "lcp create {} host-if {}{}{}".format(
                self._get_sw_if_index(configuration_item.interface),
                configuration_item.mirror_interface_host_name,
                (
                    f" netns {configuration_item.host_namespace}"
                    if configuration_item.host_namespace
                    else ""
                ),
                (
                    " tun"
                    if configuration_item.mirror_interface_type
                    == LCPHostInterfaceTypeEnum.TUN
                    else ""
                ),
            ).strip()
        else:
            cmd = "lcp delete {}".format(
                self._get_sw_if_index(configuration_item.interface)
            )
        reply = self.vpp_client.api.cli_inband(cmd=cmd)
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error="Failed to create lcp pair."
                + get_vpp_error_log_message(reply.retval),
            )

        if is_add:
            dump = self.dump_from_vpp(configuration_item)
            if not isinstance(dump, LCPPairDetail):
                logger.warning(
                    "can't retrieve host_sw_if_index (VPP interface part of tunneling from VPP to linux networking "
                    "namespace (LCP interface mirroring)): can't retrieve lcp pair listing"
                )
                return ApplyReply(success=True)
            return ApplyReply(
                success=True, vpp_data={HOST_SW_IF_INDEX: dump.host_sw_if_index}
            )
        else:
            return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        if isinstance(configuration_item, LCPPairConfigurationItem):
            reply = self.vpp_client.api.lcp_itf_pair_get()
            try:  # the binary api is a mess in VPP 22.02 (fixed in newer VPP?), so to prevent unexpected things i just use try/except # TODO still problem with currently used VPP 22.10.1 ?
                if reply[0].retval != 0:
                    logger.warning("can't retrieve lcp pairs")
                    return dict()

                phy_sw_if_index = self._get_sw_if_index(configuration_item.interface)
                for pair_detail in reply[1]:
                    if (
                        pair_detail.host_if_name
                        == configuration_item.mirror_interface_host_name
                        and pair_detail.phy_sw_if_index == phy_sw_if_index
                    ):
                        return LCPPairDetail.get_from_basic_vpp_client_equivalent_obj(
                            pair_detail
                        )
            except Exception:
                logger.exception("can't handle retrieval of lcp pairs")
            return dict()

        # the binary api for linux-cp's default namespage getting is broken too
        # reply = self.vpp_client.api.lcp_default_ns_get()
        return dict()


class FRRAgentClient(Proxy):
    def __init__(self, frr_agent_host: str = "0.0.0.0", frr_agent_port: int = 9999):
        """Creates and initializes FRR Agent client

        :param frr_agent_host: hostname/ip address of the FRR Agent's Pyro5 nameserver
        :param frr_agent_port: port of the FRR Agent's Pyro5 nameserver
        """
        serializers.import_pyro5_serializers()  # fix for serializing of custom types
        super().__init__(f"PYRONAME:frr.api@{frr_agent_host}:{frr_agent_port}")


class FRRHandler(ConfigurationItemHandler):
    def __init__(self, frr_agent_host, frr_agent_port):
        self._frr_agent_host = frr_agent_host
        self._frr_agent_port = frr_agent_port

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, FRRConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        with FRRAgentClient(
            frr_agent_host=self._frr_agent_host, frr_agent_port=self._frr_agent_port
        ) as client:
            reply = client.load_vtysh_configuration(configuration_item.config)
            if reply.return_code:
                ApplyReply(success=False, error="return code " + reply.return_code)
        return ApplyReply(success=True)

    def remove_from_vpp(self, configuration_item):
        return ApplyReply(
            success=True
        )  # do nothing, only by applying new config to FRR changes something inside FRR

    def dump_from_vpp(self, configuration_item):
        with FRRAgentClient(
            frr_agent_host=self._frr_agent_host, frr_agent_port=self._frr_agent_port
        ) as client:
            reply = client.run_vtysh_command("write terminal")
            if reply.return_code:
                FRRConfigurationItem()
            start_index = reply.stdout.find("!")
            if start_index < 0:
                start_index = 0
            end_index = reply.stdout.rfind("end")
            if end_index < 0:
                end_index = len(reply.stdout)
            return FRRConfigurationItem(config=reply.stdout[start_index:end_index])


class UnavailableDHCPSocketError(Exception):
    pass


class DHCPSocketConnetionError(Exception):
    pass


class DHCPRequestSendError(Exception):
    pass


class DHCPResponceReceiveError(Exception):
    pass


class DHCPResponceParsingError(Exception):
    pass


class DHCPHandler(ConfigurationItemHandler):

    def __init__(self, dhcp_socket):
        self._dhcp_socket = dhcp_socket

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, DHCPConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # not used by anyone

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        config = configuration_item.config

        # removing comments (not valid for json)
        if configuration_item.strip_comment_lines:
            config = "\n".join(line for line in config.splitlines() if "//" not in line)

        # local config validation
        try:
            config_dict = json.loads(config)
        except Exception as e:
            return ApplyReply(
                success=False,
                error="Config is not parsable json:"
                + repr(traceback.format_exception(e)),
            )

        # DHCP server restart (technically it is shutdown, but supervisord will start it again in docker!)
        # https://kea.readthedocs.io/en/kea-2.4.0/arm/ctrl-channel.html#the-shutdown-command
        # TODO test
        if configuration_item.restart_before_config_apply:
            try:
                response = self._send_request(
                    KeaConfigRequest(
                        command="shutdown",
                        arguments=dict(),
                    )
                )
            except Exception as e:
                return ApplyReply(
                    success=False,
                    error="Error while doing shutdown on DHCP server (part of restart):"
                    + repr(traceback.format_exception(e)),
                )
            if response.result != 0:
                return ApplyReply(
                    success=False,
                    error="DHCP server shutdown command failed (shutdown is part of restart):"
                    + repr(response.text),
                )
            # TODO some active checking of correct starting and if not detected in some time then error to client
            time.sleep(5)

        # remote config validation
        # (see https://kea.readthedocs.io/en/kea-2.4.0/arm/ctrl-channel.html#the-config-test-command)
        try:
            response = self._send_request(
                KeaConfigRequest(
                    command="config-test",
                    arguments=config_dict,
                )
            )
        except Exception as e:
            return ApplyReply(
                success=False,
                error="Error while testing config by DHCP server:"
                + repr(traceback.format_exception(e)),
            )
        if response.result != 0:
            return ApplyReply(
                success=False,
                error="Remote validation of config failed:" + repr(response.text),
            )
        # configure DHCP server with the checked config
        # (see https://kea.readthedocs.io/en/kea-2.4.0/arm/ctrl-channel.html#the-config-set-command)
        try:
            response = self._send_request(
                KeaConfigRequest(
                    command="config-set",
                    arguments=config_dict,
                )
            )
        except Exception as e:
            return ApplyReply(
                success=False,
                error="Error while applying config:"
                + repr(traceback.format_exception(e)),
            )
        if response.result != 0:
            return ApplyReply(
                success=False,
                error="Failed to set new configuration due to:" + repr(response.text),
            )
        return ApplyReply(success=True)

    def remove_from_vpp(self, configuration_item):
        return ApplyReply(
            success=True
        )  # do nothing, only by applying new config to DHCP changes something inside DHCP

    def dump_from_vpp(self, configuration_item):
        # configure DHCP server with the checked config
        # (see https://kea.readthedocs.io/en/kea-2.4.0/arm/ctrl-channel.html#the-config-get-command)
        try:
            response = self._send_request(
                KeaConfigRequest(
                    command="config-get",
                )
            )
        except Exception:
            logger.exception("Error while getting DHCP config")
            return dict()
        if response.result != 0:
            logger.warning(
                "Failed to get new DHCP configuration due to:" + repr(response.text)
            )
            return dict()
        return DHCPConfigurationItem(config=json.dumps(response.arguments))

    def _send_request(self, request: KeaConfigRequest) -> KeaConfigResponse:
        """Send request to KEA DHCP server using unix socket."""

        def recvall(socket):
            """Receive all data from unix socket"""
            data = []
            while True:
                chunk = socket.recv(4096)
                if not chunk:
                    break
                data.append(chunk)
            return b"".join(data)

        def wait_for_writable_dhcp_socket(timeout: int) -> bool:
            """Timeouted wait for DHCP socket file to be present and writable."""
            for _ in range(timeout * 2):
                if os.path.exists(self._dhcp_socket) and os.access(
                    self._dhcp_socket, os.W_OK
                ):
                    return True
                time.sleep(0.5)
            return False

        if not wait_for_writable_dhcp_socket(5):
            raise UnavailableDHCPSocketError(
                f"can't find or write DHCP control socket ({self._dhcp_socket}) within timeout "
            )

        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            try:
                client.connect(self._dhcp_socket)
            except Exception as e:
                raise DHCPSocketConnetionError(self._dhcp_socket) from e

            try:
                req_encoded = request.to_json().encode()
            except Exception as e:
                raise ValueError("can't convert request to json and encode it") from e

            try:
                client.sendall(req_encoded)
            except Exception as e:
                raise DHCPRequestSendError() from e

            try:
                response = recvall(client)
            except Exception as e:
                raise DHCPResponceReceiveError() from e

            try:
                return KeaConfigResponse.from_json(response.decode())
            except Exception as e:
                raise DHCPResponceParsingError() from e
        finally:
            # Close the connection
            client.close()


class RunningPacketCaptureHandler(ConfigurationItemHandler):
    def __init__(self, vpp_client, get_sw_if_index, get_current_running_packet_capture):
        self.vpp_client = vpp_client
        self._get_sw_if_index = get_sw_if_index
        self._get_current_running_packet_capture = get_current_running_packet_capture

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, RunningPacketCapture)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # noone uses it

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        if configuration_item.capture_only_from_interface != PCAP_ANY_INTERFACE:
            return {f"interface/{configuration_item.capture_only_from_interface}"}
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        """Starts the packet capturing defined in configuration_item"""

        # check for bad case of running multiple packet capturing at the same time
        if self._get_current_running_packet_capture() is not None:
            return ApplyReply(
                success=False,
                error=f"There can be only one packet capturing running at the same time. The current "
                f"packet capturing running is {self._get_current_running_packet_capture()}",
            )

        # configure advanced filtering
        filter_cmds = []
        for rule in configuration_item.capture_by_matching_rules:
            match rule:
                case L2PacketMatchingRule(
                    src_physical_address=src, dest_physical_address=dst, protocol=proto
                ):
                    if not src and not dst and proto == NO_PROTOCOL:  # no matching
                        continue
                    filter_cmds.append(
                        "classify filter pcap mask l2 {}{}{}match l2{}{}{}".format(
                            "src " if src else "",
                            "dst " if dst else "",
                            "proto " if proto else "",
                            f" src {src}" if src else "",
                            f" dst {dst}" if dst else "",
                            f" proto {proto}" if proto else "",
                        )
                    )
                case L3PacketMatchingRule(
                    src_ipv4_address=src, dest_ipv4_address=dst, protocol=proto
                ):
                    if not src and not dst and proto == NO_PROTOCOL:  # no matching
                        continue
                    filter_cmds.append(
                        "classify filter pcap mask l3 ip4 {}{}{}match l3 ip4{}{}{}".format(
                            "src " if src else "",
                            "dst " if dst else "",
                            "proto " if proto else "",
                            f" src {src}" if src else "",
                            f" dst {dst}" if dst else "",
                            f" proto {proto}" if proto else "",
                        )
                    )
                case _:
                    logger.warning(f"Ignoring unknown packet matching rule: {rule}")
        if filter_cmds:
            reply = self.vpp_client.api.cli_inband(cmd="classify filter pcap del")
            if reply.retval != 0:
                return ApplyReply(
                    success=False,
                    error="Failed to clean classify filters."
                    + get_vpp_error_log_message(reply.retval),
                )
            for cmd in filter_cmds:
                reply = self.vpp_client.api.cli_inband(cmd=cmd)
                if reply.retval != 0:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to apply one of classify filters('{cmd}'). "
                        f"{get_vpp_error_log_message(reply.retval)}(CLI reply: {reply.reply})",
                    )

        # configure pcap capturing start
        interface_name = configuration_item.capture_only_from_interface
        if interface_name != PCAP_ANY_INTERFACE:
            sw_if_index = self._get_sw_if_index(interface_name)  # logical name to index
            interfaces = self.vpp_client.api.sw_interface_dump(sw_if_index=sw_if_index)
            if not interfaces:
                return ApplyReply(
                    success=False,
                    error=f"Failed to resolve VPP name for interface {interface_name}). "
                    + get_vpp_error_log_message(reply.retval),
                )
            interface_name = interfaces[0].interface_name  # get VPP name of interface

        pcap_start = "pcap trace {}{}{}max {} intfc {} file {} max-bytes-per-pkt {}{}{}{}".format(
            "rx " if configuration_item.capture_rx_packets else "",
            "tx " if configuration_item.capture_tx_packets else "",
            "drop " if configuration_item.capture_dropped_packets else "",
            configuration_item.max_packets_to_capture,
            interface_name,
            "{}-{}.pcap".format(
                configuration_item.output_file_prefix,
                datetime.now().strftime("%Y-%m-%d--%H-%M-%S-%f"),
            ),
            configuration_item.max_bytes_per_packet,
            " filter" if filter_cmds else "",
            " preallocate-data" if configuration_item.preallocate_data_buffer else "",
            " free-data" if configuration_item.free_data_buffer else "",
        )
        reply = self.vpp_client.api.cli_inband(cmd=pcap_start)
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error=f"Failed to start pcap capturing (CLI command: {pcap_start}). "
                f"{get_vpp_error_log_message(reply.retval)}(CLI reply: {reply.reply})",
            )

        return ApplyReply(success=True)

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        """Stops the packet capturing defined in configuration_item"""

        # check for bad case of stopping the wrong capturing
        if self._get_current_running_packet_capture() is not configuration_item:
            return ApplyReply(
                success=False,
                error=f"Only packet capturing that is currently running can be stopped (currently "
                f"running: {self._get_current_running_packet_capture()}, wanted to stop: "
                f"{configuration_item})",
            )

        # stop tracing
        reply = self.vpp_client.api.cli_inband(cmd="pcap trace off")
        if reply.retval != 0 and "No packets captured" not in reply.reply:
            return ApplyReply(
                success=False,
                error=f"Failed to stop pcap tracing. {get_vpp_error_log_message(reply.retval)}"
                f"(CLI reply: {reply.reply})",
            )

        # extract info from tracing stop
        captured_packets_count = 0
        filename = ""
        if "No packets captured" not in reply.reply:
            match re.findall(r"(\d*) packets to (\/[^,]*)", reply.reply):
                case [(packets, file)]:
                    try:
                        captured_packets_count = int(packets)
                    except ValueError:
                        return ApplyReply(
                            success=False,
                            error=f"Failed to parse packet count from string '{packets}'",
                        )
                    filename = file
                case _:
                    return ApplyReply(
                        success=False,
                        error=f"Failed to parse packet capturing stop output '{reply.reply}'",
                    )

        # cleanup advanced filtering
        reply = self.vpp_client.api.cli_inband(cmd="classify filter pcap del")
        if reply.retval != 0:
            return ApplyReply(
                success=False,
                error=f"Failed to clean classify filters. "
                f"{get_vpp_error_log_message(reply.retval)}(CLI reply: {reply.reply})",
            )

        return ApplyReply(
            success=True,
            vpp_data={
                "packet_capture_output_file": filename,
                "count_of_captured_packets": captured_packets_count,
            },
        )

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        return dict()


class LinuxDockerVethTunnelHandler(ConfigurationItemHandler):
    def __init__(self):
        pass

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, LinuxDockerVethTunnelConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # noone uses it

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        if not configuration_item.docker_container_name:
            return ApplyReply(success=False, error="Container name must be filled")

        # prepare docker container (network namespace linking)
        try:
            container_pid = subprocess.check_output(
                "docker inspect -f '{{.State.Pid}}' "
                + configuration_item.docker_container_name,
                shell=True,
            ).decode()
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to get docker container id for docker container "
                f"{configuration_item.docker_container_name}. Error code {e.returncode} ({e})",
            )
        container_pid = container_pid.replace("\n", "")
        try:
            subprocess.check_call(
                f"sudo mkdir -p /var/run/netns;sudo ln -sf /proc/{container_pid}/ns/net "
                f'"/var/run/netns/{configuration_item.docker_container_name}"',
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to link docker network namespace into standard linux network namespace "
                f"directory. Docker container name: {configuration_item.docker_container_name}. "
                f"Error code {e.returncode} ({e})",
            )

        # create veth pair and position one end into docker container
        try:
            subprocess.check_call(
                f"sudo ip link add {configuration_item.host_interface_name} type veth "
                f"peer name {configuration_item.docker_interface_name}",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to create veth tunnel. Error code {e.returncode} ({e})",
            )

        try:
            subprocess.check_call(
                f"sudo ip link set {configuration_item.docker_interface_name} netns "
                f"{configuration_item.docker_container_name}",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to move one end of veth tunnel into dokcer container "
                f"{configuration_item.docker_container_name}. Error code {e.returncode} ({e})",
            )
        # set interface up
        try:
            subprocess.check_call(
                f"sudo ip netns exec {configuration_item.docker_container_name} ip link set "
                f"{configuration_item.docker_interface_name} up",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to enable docker end of veth tunnel. Error code {e.returncode} ({e})",
            )
        try:
            subprocess.check_call(
                f"sudo ip link set {configuration_item.host_interface_name} up",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to enable host end of veth tunnel. Error code {e.returncode} ({e})",
            )

        # configure ip addresses for both veth tunnel endpoints
        for ip_address in configuration_item.host_interface_ip_addresses:
            try:
                subprocess.check_call(
                    f"sudo ip addr add {ip_address} dev {configuration_item.host_interface_name}",
                    shell=True,
                )
            except subprocess.CalledProcessError as e:
                return ApplyReply(
                    success=False,
                    error=f"Failed to set address {ip_address} to host end of veth tunnel. "
                    f"Error code {e.returncode} ({e})",
                )
        for ip_address in configuration_item.docker_interface_ip_addresses:
            try:
                subprocess.check_call(
                    f"sudo ip netns exec {configuration_item.docker_container_name} ip addr add {ip_address} "
                    f"dev {configuration_item.docker_interface_name}",
                    shell=True,
                )
            except subprocess.CalledProcessError as e:
                return ApplyReply(
                    success=False,
                    error=f"Failed to set address {ip_address} to docker end of veth tunnel. "
                    f"Error code {e.returncode} ({e})",
                )

        return ApplyReply(success=True)

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        # disable veth tunnel ends
        try:
            subprocess.check_call(
                f"sudo ip netns exec {configuration_item.docker_container_name} ip link set "
                f"{configuration_item.docker_interface_name} down",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to disable docker end of veth tunnel. Error code {e.returncode} ({e})",
            )
        try:
            subprocess.check_call(
                f"sudo ip link set {configuration_item.host_interface_name} down",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to disable host end of veth tunnel. Error code {e.returncode} ({e})",
            )
        # remove veth tunnel
        try:
            subprocess.check_call(
                f"sudo ip link delete {configuration_item.host_interface_name}",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to remove veth tunnel. Error code {e.returncode} ({e})",
            )

        # unlink docker container network namespace
        try:
            subprocess.check_call(
                f"sudo unlink /var/run/netns/{configuration_item.docker_container_name}",
                shell=True,
            )
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to unlink docker container network namespace. "
                f"Error code {e.returncode} ({e})",
            )

        return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        return dict()


class LinuxRouteHandler(ConfigurationItemHandler):
    def __init__(self):
        pass

    def is_handling_configuration_item(self, configuration_item):
        return isinstance(configuration_item, LinuxRouteConfigurationItem)

    @validate_config_item
    def get_labels(self, configuration_item):
        return set()  # noone uses it

    @validate_config_item
    def get_dependency_labels(self, configuration_item):
        return set()

    @validate_config_item
    def add_to_vpp(self, configuration_item):
        return self.add_remove_route(configuration_item, is_add=True)

    @validate_config_item
    def remove_from_vpp(self, configuration_item):
        return self.add_remove_route(configuration_item, is_add=False)

    def add_remove_route(self, configuration_item, is_add=True):
        add_or_remove = "add" if is_add else "del"
        basecmd = "{}ip route {} {}".format(
            (
                f"docker exec {configuration_item.docker_container_name} "
                if configuration_item.docker_container_name
                else "sudo "
            ),
            add_or_remove,
            configuration_item.destination_network,
        )
        via = (
            "via " + configuration_item.next_hop_address
            if configuration_item.next_hop_address
            else ""
        )
        dev = (
            "dev " + configuration_item.outgoing_interface
            if configuration_item.outgoing_interface
            else ""
        )
        cmd = " ".join([basecmd, via, dev])
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError as e:
            return ApplyReply(
                success=False,
                error=f"Failed to {add_or_remove} linux route (cmd='{cmd}'). "
                f"Error code {e.returncode} ({e})",
            )

        return ApplyReply(success=True)

    @validate_config_item
    def dump_from_vpp(self, configuration_item):
        return dict()



"""
Custom field types used in modeled API(dataclasses) so that the Pyro5 framework can serialize the content correctly
"""

import ipaddress

from marshmallow.fields import Field


class IPNetwork(Field):
    """A IPNetwork field."""

    default_error_messages = {"invalid_ip_network": "Not a valid IP network."}

    DESERIALIZATION_CLASS = None  # type: typing.Optional[typing.Type]

    def __init__(self, *args, exploded: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.exploded = exploded

    def _serialize(self, value, attr, obj, **kwargs) -> str | None:
        if value is None:
            return None
        if self.exploded:
            return value.exploded
        return value.compressed

    def _deserialize(
        self, value, attr, data, **kwargs
    ) -> None | (ipaddress.IPv4Network | ipaddress.IPv6Network):
        def ensure_text_type(val: str | bytes) -> str:
            if isinstance(val, bytes):
                val = val.decode("utf-8")
            return str(val)

        if value is None:
            return None
        try:
            return (self.DESERIALIZATION_CLASS or ipaddress.ip_network)(
                ensure_text_type(value)
            )
        except (ValueError, TypeError) as error:
            raise self.make_error("invalid_ip_network") from error


class IPv4Network(IPNetwork):
    """A IPv4 Network field."""

    default_error_messages = {"invalid_ip_network": "Not a valid IPv4 network."}

    DESERIALIZATION_CLASS = ipaddress.IPv4Network


class IPv6Network(IPNetwork):
    """A IPv6 Network field."""

    default_error_messages = {"invalid_ip_network": "Not a valid IPv6 network."}

    DESERIALIZATION_CLASS = ipaddress.IPv6Network



"""
All the model data classes and help constants/structures for public/inner API
"""

import base64
import ipaddress
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto, IntEnum
from typing import Dict, List, Any, Union

from dataclasses_json import dataclass_json, config
from marshmallow import fields

from vrouter.common import custom_marshmallow_fields

#################################
### Constants, variables
#################################

NO_INTERFACE_SW_INDEX = 0xFFFFFFFF
NO_ACL_INDEX = 0xFFFFFFFF
NO_WIREGUARD_PEER_INDEX = 0xFFFFFFFF
NOT_SET_MTU = -1
SW_IF_INDEX = "sw_if_index"
HOST_SW_IF_INDEX = "host_sw_if_index"
ACL_INDEX = "acl_index"
WIREGUARD_PEER_INDEX = "peer_index"
VPP_DUMP_KEY = "vpp_dump"
PCAP_ANY_INTERFACE = "any"
NO_PROTOCOL = -1

# Note: There is some problem with dynamic loading of "vl_api_if_status_flags_t" in VppEnum -> redefining constants
# from VPP API json files

IF_STATUS_API_FLAG_ADMIN_UP = (
    1  # VppEnum.vl_api_if_status_flags_t.IF_STATUS_API_FLAG_ADMIN_UP
)
IF_STATUS_API_FLAG_LINK_UP = (
    2  # VppEnum.vl_api_if_status_flags_t.IF_STATUS_API_FLAG_LINK_UP
)


class FibPathType:
    FIB_PATH_TYPE_NORMAL = 0
    FIB_PATH_TYPE_LOCAL = 1
    FIB_PATH_TYPE_DROP = 2
    FIB_PATH_TYPE_UDP_ENCAP = 3
    FIB_PATH_TYPE_BIER_IMP = 4
    FIB_PATH_TYPE_ICMP_UNREACH = 5
    FIB_PATH_TYPE_ICMP_PROHIBIT = 6
    FIB_PATH_TYPE_SOURCE_LOOKUP = 7
    FIB_PATH_TYPE_DVR = 8
    FIB_PATH_TYPE_INTERFACE_RX = 9
    FIB_PATH_TYPE_CLASSIFY = 10


class FibPathFlags:
    FIB_PATH_FLAG_NONE = 0
    FIB_PATH_FLAG_RESOLVE_VIA_ATTACHED = 1
    FIB_PATH_FLAG_RESOLVE_VIA_HOST = 2
    FIB_PATH_FLAG_POP_PW_CW = 4


class FibPathProto:
    FIB_PATH_NH_PROTO_IP4 = 0
    FIB_PATH_NH_PROTO_IP6 = 1
    FIB_PATH_NH_PROTO_MPLS = 2
    FIB_PATH_NH_PROTO_ETHERNET = 3
    FIB_PATH_NH_PROTO_BIER = 4
    FIB_PATH_NH_PROTO_NSH = 5


#################################
### Configuration models
#################################


class ConfigurationItem(ABC):
    """Abstraction of the smallest piece of configuration that can be applied to VPP/FRR/Linux namespace"""

    pass


class ConfigurationItemHandler(ABC):
    """Performs functionality related to one given class of ConfigurationItem.

    Covers interaction with VPP, dependency handling. It is inspired by descriptors in Ligato VPP-Agent repository
    (https://github.com/ligato/vpp-agent/blob/6c38360edf8285915202c8e4192e77a5efd71307/plugins/kvscheduler/api/kv_descriptor_api.go#L133)
    """

    @abstractmethod
    def add_to_vpp(self, configuration_item):
        """Applies configuration represented by configuration_item into VPP (or FRR or Linux namespace)"""
        pass

    @abstractmethod
    def remove_from_vpp(self, configuration_item):
        """Removes configuration represented by configuration_item from VPP (or FRR or Linux namespace)"""
        pass

    @abstractmethod
    def dump_from_vpp(self, configuration_item):
        """Retrieves configuration from VPP/FRR/Linux related to configuration represented by configuration_item"""
        pass

    @abstractmethod
    def is_handling_configuration_item(self, configuration_item):
        """Checks whenever this ConfigurationItemHandler handles given configuration_item. It can be used to find
        the correct ConfigurationItemHandler for given configuration_item or the guard method calls of given
        ConfigurationItemHandler against unexpected configuration_item input."""
        pass

    @abstractmethod
    def get_labels(self, configuration_item):
        """Provide labels for given configuration_item. Their presence represent correct application
        of the configuration_item into the VPP/FRR/Linux namespace. By referencing these labels, dependency connections
        can be established between configuration and correct ordering of configuration items can be achieved.
        (See get_dependency_labels(...))"""
        pass

    @abstractmethod
    def get_dependency_labels(self, configuration_item):
        """Provides labels of configuration items that given configuration item depends on. (see get_labels(...))"""
        pass


class VPPConfigUpdater(ABC):
    """Performs updates of existing ConfigurationItems already applied in VPP.

    This class should be used as alternative to delete correctly applied ConfigurationItem in VPP and that adding
    updated Configuration back to VPP. However, its main purpose is the make only necessary changes in VPP in compare
    to entire delete of entity and recreating it from scratch (i.e. update ip address of interface without deleting it
    and recreating it with new ip address)"""

    @abstractmethod
    def update_in_vpp(self, old_configuration_item, new_configuration_item):
        pass


class State(Enum):
    """State of configuration item that is sent to server from user using API."""

    RECEIVED = auto()
    """Just received, not processed yet. Initial state"""
    BLOCKED = auto()
    """Applying of configuration item to VPP/FRR/Linux namespace is blocked due to configuration dependency that 
    is not met yet. Waiting until all the dependencies are satisfied and only then trying to apply it."""
    BLOCKED_BY_ADD_APPLY_FAILURE = auto()
    """Similar as BLOCKED, but the reason is for blocking further action is failure/error from applying(adding) 
    to VPP/FRR/Linux namespace"""
    BLOCKED_BY_REMOVE_APPLY_FAILURE = auto()
    """Similar as BLOCKED, but the reason is for blocking further action is failure/error from applying(removing) 
    to VPP/FRR/Linux namespace"""
    BLOCKED_BY_UPDATE_APPLY_FAILURE = auto()
    """Similar as BLOCKED, but the reason is for blocking further action is failure/error from applying(updating) 
    to VPP/FRR/Linux namespace"""
    APPLIED = auto()
    """Configuration item is successfully applied to VPP/FRR/Linux namespace."""


class InterfaceType(Enum):
    UNDEFINED_TYPE = 0
    SUB_INTERFACE = 1
    SOFTWARE_LOOPBACK = 2
    DPDK = 3
    MEMIF = 4
    TAP = 5
    AF_PACKET = 6
    VXLAN_TUNNEL = 7
    IPSEC_TUNNEL = (
        8  # Deprecated in VPP 20.01 +.Use IPIP_TUNNEL + ipsec.TunnelProtection instead.
    )
    VMXNET3_INTERFACE = 9
    BOND_INTERFACE = 10
    GRE_TUNNEL = 11
    GTPU_TUNNEL = 12
    IPIP_TUNNEL = 13
    WIREGUARD_TUNNEL = 14
    RDMA = 15

    @staticmethod
    def get_from_vpp_str(vpp_if_type_str):
        match vpp_if_type_str:
            case "RDMA interface":
                return InterfaceType.RDMA
            case "local" | "loop":
                return InterfaceType.SOFTWARE_LOOPBACK
            case "memif":
                return InterfaceType.MEMIF
            case "tap" | "tun":
                return InterfaceType.TAP
            case "host":
                return InterfaceType.AF_PACKET
            case "vxlan":
                return InterfaceType.VXLAN_TUNNEL
            case "ipsec":
                return InterfaceType.IPSEC_TUNNEL
            case "vmxnet3":
                return InterfaceType.VMXNET3_INTERFACE
            case "Bond":
                return InterfaceType.BOND_INTERFACE
            case "gtpu":
                return InterfaceType.GTPU_TUNNEL
            case "ipip":
                return InterfaceType.IPIP_TUNNEL
            case "wireguard":
                return InterfaceType.WIREGUARD_TUNNEL
            case "dpdk":
                return InterfaceType.DPDK
            case "virtio":
                return (
                    InterfaceType.TAP
                )  # FIXME there could be non-tap virtio interfaces
            case _:
                return InterfaceType.DPDK


class GRELinkType(Enum):
    L3 = 0
    """L3 GRE - the tunnel is in L3 mode"""
    TEB = 1
    """Transaparent Ethernet Bridging - the tunnel is in L2 mode"""
    ERSPAN = 2
    """ERSPAN - the tunnel is for port mirror SPAN output"""


@dataclass_json
@dataclass
class GREInterfaceLink:
    type: GRELinkType = GRELinkType.L3
    """Type of GRE tunnel"""
    src_addr: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Source IP address"""
    dst_addr: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Destination IP address"""
    outer_fib_id: int = 0
    """Encapsulation FIB table ID"""
    session_id: int = 0
    """Session for tunnel of ERSPAN type, range 0-1023"""


@dataclass_json
@dataclass
class WireguardInterfaceLink:
    private_key: str = ""
    """Private-key encoded in base64"""
    port: int = 0
    """Listen UDP port"""
    src_addr: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Source IP address"""


@dataclass_json
@dataclass
class InterfaceConfigurationItem(ConfigurationItem):
    name: str
    type: InterfaceType = InterfaceType.TAP
    enabled: bool = True
    ip_addresses: List[str] = field(default_factory=lambda: [])
    mtu: int = NOT_SET_MTU
    """ MTU (Maximum Transmission Unit) of this interface"""
    # phys_address: str = '' # not used -> ignored
    link: Union[GREInterfaceLink, WireguardInterfaceLink, None] = None
    """Link defines configuration for specific interface types."""


class RouteType(Enum):
    INTRA_VRF = 0
    """Forwarding is being done in the specified vrf_id only, or according to
       the specified outgoing interface."""
    # INTER_VRF = 1
    # """ Forwarding is being done by lookup into a different VRF,
    #     specified as via_vrf_id field. In case of these routes, the outgoing
    #     interface should not be specified. The next hop IP address
    #     does not have to be specified either, in that case VPP does full
    #     recursive lookup in the via_vrf_id VRF."""
    DROP = 2
    """ Drops the network communication designated for specific IP address."""


@dataclass_json
@dataclass
class RouteOutputPath:
    """Output path part of route. It should be used to define multiple output possibilities for route. That means when
    you want to steer some packets for given network prefix into route loadbalancer that uses multiple outputs (i.e.
    multiple wg tunnels)"""

    next_hop_address: str = ""
    """ Next hop ip address """
    outgoing_interface: str = ""
    """ Name of outgoing interface that should be used for this route. """
    weight: int = 1
    """ Weight is used for unequal cost load balancing (uint32 in vpp)."""  # OSPF learned paths can use this


@dataclass_json
@dataclass
class RouteConfigurationItem(ConfigurationItem):
    type: RouteType = RouteType.INTRA_VRF
    """ Type of route defining route behaviour. See RouteType."""
    destination_network: str = ""
    """ Destination ip network for steering traffic using this route. 
    It has format <ip address>/<prefix> (i.e. "10.1.2.0/24)."""
    next_hop_address: str = ""
    """ Next hop ip address """
    outgoing_interface: str = ""
    """ Name of outgoing interface that should be used for this route. """
    weight: int = 1
    """ Weight is used for unequal cost load balancing (uint32 in vpp)."""  # OSPF learned paths can use this
    multi_output_paths: List[RouteOutputPath] = field(default_factory=lambda: [])
    """ Definition for multiple output paths for route. If output path is only one, use the direct fields 
    in RouteConfigurationItem. When this field is used, the direct duplicated direct fields in RouteConfigurationItem 
    will be ignored and only multi_output_paths will be used."""
    # NOT USED or MEANINGLESS for given usecases:
    #
    # vrf_id: int = 0 # ignored as all routes should use vrf 0
    #
    # preference:int
    # """ Preference defines path preference. Lower preference is preferred.
    # Only paths with the best preference contribute to forwarding (a poor man's primary and backup).
    # It is uint32 in vpp."""
    #
    # via_vrf_id: int
    # """ Specifies VRF ID for the next hop lookup / recursive lookup. It is uint32 in vpp."""


class ACLAction(Enum):
    DENY = 0
    PERMIT = 1
    REFLECT = 2


@dataclass_json
@dataclass
class ICMPRange:
    """Range for ICMP codes and types
    See https://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml"""

    first: int = 0  # minimum: 0 maximum: 255
    last: int = 255  # minimum: 0 maximum: 255


@dataclass_json
@dataclass
class ICMPProtocol:
    icmpv6: bool = False
    """ICMPv6 flag, if false ICMPv4 will be used"""
    icmp_code_range: ICMPRange = field(default_factory=lambda: ICMPRange())
    """Inclusive range representing icmp codes to be used"""
    icmp_type_range: ICMPRange = field(default_factory=lambda: ICMPRange())
    """Inclusive range representing icmp types to be used"""


@dataclass_json
@dataclass
class PortRange:
    """Inclusive range representing destination ports to be used.
    When only lower-port is present, it represents a single port."""

    lower_port: int = 0  # minimum: 0 maximum: 65535
    upper_port: int = 65535  # minimum: 0 maximum: 65535


@dataclass_json
@dataclass
class TCPProtocol:
    destination_port_range: PortRange = field(default_factory=lambda: PortRange())
    source_port_range: PortRange = field(default_factory=lambda: PortRange())
    tcp_flags_mask: int = 0  # minimum: 0 maximum: 255
    """ Binary mask for tcp flags to match. MSB order (FIN at position 0).
    Applied as logical AND to tcp flags field of the packet being matched,
    before it is compared with tcp-flags-value.
    """
    tcp_flags_value: int = 0  # minimum: 0 maximum: 255
    """ Binary value for tcp flags to match. MSB order (FIN at position 0).
    Before tcp-flags-value is compared with tcp flags field of the packet being matched,
    tcp-flags-mask is applied to packet field value.
    """


@dataclass_json
@dataclass
class UDPProtocol:
    destination_port_range: PortRange = field(default_factory=lambda: PortRange())
    source_port_range: PortRange = field(default_factory=lambda: PortRange())


@dataclass_json
@dataclass
class OtherProtocol:
    """ACL rule's packet selector based on IP protocol.
    For ICMP,TCP,UDP protocol selection use custome designed packet selectors (ICMPProtocol,TCPProtocol,UDPProtocol).
    """

    protocol: int = 0
    """ IP protocol number (http://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml) defining protocol 
    filtering for ACL rules"""


@dataclass_json
@dataclass
class IPAddresses:
    destination_network: ipaddress.IPv4Network = field(
        default_factory=lambda: ipaddress.IPv4Network("0.0.0.0/0")
    )  # FIXME no support for IPv6 in ACL
    """ Destination IPv4/IPv6 network address (<ip>/<network>) """
    source_network: ipaddress.IPv4Network = field(
        default_factory=lambda: ipaddress.IPv4Network("0.0.0.0/0")
    )
    """ Destination IPv4/IPv6 network address (<ip>/<network>) """


@dataclass_json
@dataclass
class IPSpecification:
    addresses: Union[IPAddresses, None] = None
    protocol: Union[ICMPProtocol, TCPProtocol, UDPProtocol, OtherProtocol, None] = None


@dataclass_json
@dataclass
class ACLRuleConfigurationItem:
    action: ACLAction = ACLAction.DENY
    refinement: Union[IPSpecification, None] = None
    # Note: ignoring MAC ACL rules(MAcSpecification) as they are not
    # needed for failover case->ACLRuleConfigurationItem=IP-based Rule


@dataclass_json
@dataclass
class ACLConfigurationItem(ConfigurationItem):
    name: str = ""
    rules: List[ACLRuleConfigurationItem] = field(
        default_factory=lambda: [ACLRuleConfigurationItem()]
    )
    egress: List[str] = field(
        default_factory=lambda: []
    )  # TODO map ACL to interfaces in implementation
    """Names of egress interfaces"""
    ingress: List[str] = field(default_factory=lambda: [])
    """Names of ingress interfaces"""


# NAT44 parts ignored from Ligato's vpp-agent nat44 proto models:
# message Nat44Global {
#     // Enable/disable forwarding. By default it is disabled.
#     bool forwarding = 1;
#     // Enable/disable endpoint-independent mode. by default disabld.
#     bool endpoint_independent = 5;
#     // Virtual reassembly for IPv4.
#     VirtualReassembly virtual_reassembly = 4;
# }


# numbers from: http://www.iana.org/assignments/protocol-numbers
# defined for VPP v22.10.1 in file src/vnet/ip/protocols.def
class ProtocolInNAT(IntEnum):
    """Available protocols that can be used in NAT configuration"""

    TCP = 6
    UDP = 17
    ICMP = 1


@dataclass_json
@dataclass
class LocalIP:
    """LocalIP defines a local IP addresses."""

    # We use only one VRF in failover case
    # vrf_id:int = 0
    # """VRF (table) ID. Non-zero VRF has to be explicitly created (see api/models/vpp/l3/vrf.proto)."""
    local_ip: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Local IP address"""
    local_port: int = 0
    """Port (do not set for address mapping)."""
    probability: int = 0
    """Probability level for load-balancing mode."""


@dataclass_json
@dataclass
class StaticMapping:
    """StaticMapping defines a list of static mappings in DNAT."""

    external_interface: str = ""
    """Interface to use external IP from; preferred over external_ip."""
    external_ip: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """External IPV4 address."""
    external_port: int = 0
    """Port (do not set for address mapping)."""
    local_ips: List[LocalIP] = field(default_factory=lambda: [])
    """List of local IP addresses. If there is more than one entry, load-balancing is enabled."""
    protocol: ProtocolInNAT = ProtocolInNAT.TCP
    """Protocol used for static mapping."""

    # Not using twice-nat in failover case
    # class TwiceNatMode(Enum):
    #     DISABLED = 0
    #     ENABLED = 1
    #     SELF = 2
    #
    # twice_nat:TwiceNatMode = TwiceNatMode.DISABLED
    # """Enable/disable (self-)twice NAT."""
    # twice_nat_pool_ip:str = ''
    # """IP address from Twice-NAT address pool that should be used as source IP in twice-NAT processing.
    #    This is override for default behaviour of choosing the first IP address from twice-NAT pool that
    #    has available at least one free port (NAT is tracking translation sessions and exhausts free ports
    #    for given IP address). This is needed for example in use cases when multiple twice-NAT
    #    translations need to use different IP Addresses as source IP addresses.
    #    This functionality works with VPP 20.09 and newer. It also needs to have twice_nat set to ENABLED.
    #    It doesn't work for load-balanced static mappings (=local_ips has multiple values)."""

    # Not used in failover case
    # session_affinity:int = 0
    # """Session affinity. 0 means disabled, otherwise client IP affinity sticky time in seconds."""


@dataclass_json
@dataclass
class IdentityMapping:
    """IdentityMapping defines an identity mapping in DNAT."""

    # We use only one VRF in failover case
    # vrf_id:int = 0
    # """VRF (table) ID. Non-zero VRF has to be explicitly created (see api/models/vpp/l3/vrf.proto)."""
    interface: str = ""
    """Name of the interface to use address from; preferred over ip_address."""
    ip_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """IPv4 address"""
    port: int = 0
    """Port (do not set for address mapping)"""
    protocol: ProtocolInNAT = ProtocolInNAT.TCP
    """Protocol used for identity mapping."""


@dataclass_json
@dataclass
class DNat44ConfigurationItem(ConfigurationItem):
    """DNat44 defines destination NAT44 configuration."""

    label: str = ""
    """Unique identifier for the DNAT configuration."""
    static_mappings: List[StaticMapping] = field(default_factory=lambda: [])
    """A list of static mappings in DNAT."""
    identity_mappings: List[IdentityMapping] = field(default_factory=lambda: [])
    """A list of identity mappings in DNAT."""


@dataclass_json
@dataclass
class Nat44InterfaceConfigurationItem(ConfigurationItem):
    """Nat44Interface defines a local network interfaces enabled for NAT44."""

    name: str = ""
    """interface name (logical)."""
    nat_inside: bool = False
    """Enable/disable NAT on inside."""
    nat_outside: bool = False
    """Enable/disable NAT on outside."""
    output_feature: bool = False
    """Enable/disable output feature."""


@dataclass_json
@dataclass
class Nat44AddressPoolConfigurationItem(ConfigurationItem):
    """Nat44AddressPool defines an address pool used for NAT44."""

    name: str = ""
    """Unique name for address pool"""
    # We use only one VRF in failover case
    # vrf_id: int = 0
    # """ VRF id of tenant, 0xFFFFFFFF means independent of VRF.
    # Non-zero (and not all-ones) VRF has to be explicitly created (see api/models/vpp/l3/vrf.proto)."""
    first_ip: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """First IP address of the pool"""
    last_ip: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Last IP address of the pool. Should be higher than first_ip or empty."""
    # Not using twice-nat in failover case
    # twice_nat: bool = False
    # """Enable/disable twice NAT."""


class WireguardPeerFlags(IntEnum):
    NO_FLAGS = 0
    STATUS_DEAD = 1
    ESTABLISHED = 2


@dataclass_json
@dataclass
class WireguardPeerConfigurationItem(ConfigurationItem):
    public_key: str = ""
    """Public-key base64"""
    port: int = 0
    """Peer UDP port"""
    endpoint: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    """Endpoint IP"""
    allowed_ips: List[str] = field(default_factory=lambda: [])
    """Allowed IPs for peer. Use ip address prefixes, i.e. ["10.1.2.0/24"]. To enable all ip addresses use 
    ["0.0.0.0/0"]"""
    wg_if_name: str = ""
    """The name of the wireguard interface to which this peer belongs"""
    flags: int = int(WireguardPeerFlags.NO_FLAGS)
    """Flags. Like int(WireguardPeerFlags.STATUS_DEAD)"""
    persistent_keepalive: int = 10
    """Keepalive interval (sec)"""


@dataclass_json
@dataclass
class FRRConfigurationItem(ConfigurationItem):
    """Configuration of FRR instance"""

    config: str = ""
    """String that represent the whole FRR configuration file content."""


@dataclass_json
@dataclass
class KeaConfigRequest:
    """Request sent to KEA DHCP server."""

    command: str
    arguments: Dict = field(default_factory=lambda: (dict()))


@dataclass_json
@dataclass
class KeaConfigResponse:
    """Response from KEA DHCP server to some request(see KeaConfigRequest)"""

    result: int
    """	Status code of the request processing. Zero means success and non-zero failure (see 
    https://kea.readthedocs.io/en/kea-2.5.0/arm/ctrl-channel.html#data-syntax)
    """

    text: str = ""
    """ Description of the occurred error """

    arguments: Dict = field(default_factory=lambda: (dict()))
    """ Additional data, i.e. whole config in case of "config-get" command in request """


@dataclass_json
@dataclass
class DHCPConfigurationItem(ConfigurationItem):
    """Configuration of DHCP instance"""

    config: str = ""
    """String that represent the whole DHCP configuration file content."""

    restart_before_config_apply: bool = False
    """Switch for restarting the DHCP before applying the config field. This is workaround for DHCP server not able to 
    pickup interface created after its start. This is false by default, because normally only initial configuration 
    will need this (DHCP starts before interface creation linking it to VPP)."""

    strip_comment_lines: bool = True
    """Strips comment lines (every line with // inside) from config string before applying it to DHCP server. This is 
    needed because config input is pure JSON and this format there are not comments allowed. Comment stripping can be 
    also done on client side if needed (and turning server-side comment stripping off). """


@dataclass_json
@dataclass
class LCPGlobalsConfigurationItem(ConfigurationItem):
    """Global settings for LCP"""

    default_namespace: str = ""
    lcp_sync: bool = True
    """Enables copying of changes made in VPP into their Linux counterpart"""
    lcp_auto_subint: bool = False
    """When enabled, sub-interface creation in VPP automatically creates a Linux Interface Pair(LIP) and 
    its companion Linux network interface"""


class LCPHostInterfaceTypeEnum(IntEnum):  # vl_api_lcp_itf_host_type_t
    TAP = 0
    TUN = 1


@dataclass_json
@dataclass
class LCPPairConfigurationItem(ConfigurationItem):
    interface: str = ""
    """Interface in VPP that should be mirrored into linux namespace"""
    mirror_interface_host_name: str = ""
    """Name of host(linux) interface (host end of tap/tun tunnel from VPP) that is mirroring of VPP interface 
    (LCPPairConfigurationItem.interface)"""
    mirror_interface_type: LCPHostInterfaceTypeEnum = LCPHostInterfaceTypeEnum.TAP
    """Interface type of host(linux) interface (host end of tap/tun tunnel from VPP) that is mirroring of VPP interface 
    (LCPPairConfigurationItem.interface)"""
    host_namespace: str = ""
    """Host/Linux namespace where the VPP interface(LCPPairConfigurationItem.interface) should be mirrored"""


@dataclass_json
@dataclass
class L2PacketMatchingRule:
    src_physical_address: str = ""
    """Frame source physical address to match by. Leave empty you don't want to match frames by this."""
    dest_physical_address: str = ""
    """Frame destination physical address to match by. Leave empty you don't want to match frames by this."""
    protocol: int = NO_PROTOCOL
    """L2 protocol number to match by (see https://github.com/FDio/vpp/blob/stable/2206/src/vnet/ethernet/types.def 
    for L2 protocol numbers). Use NO_PROTOCOL to not filted by protocol. Use decimal format of protocol numbers."""


@dataclass_json
@dataclass
class L3PacketMatchingRule:
    src_ipv4_address: str = ""
    """Packet source IPv4 address to match by. Leave empty you don't want to match frames by this."""
    dest_ipv4_address: str = ""
    """Packet destination IPv4 address to match by. Leave empty you don't want to match frames by this."""
    protocol: int = NO_PROTOCOL
    """L3 protocol number to match by (see https://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml 
    for protocol numbers). Use NO_PROTOCOL to not filted by protocol. Use decimal format of protocol numbers."""


@dataclass_json
@dataclass
class RunningPacketCapture(ConfigurationItem):
    """Configuration of running packet capturing.

    By adding this configuration, packet capturing is started with given configuration settings. By removing it,
    the packet capturing is stopped and content of local packet buffer written to output file.
    Check VPP documentation for further configuration explanation:
    https://s3-docs.fd.io/vpp/22.06/cli-reference/clis/clicmd_src_vnet.html#pcap-trace,
    https://s3-docs.fd.io/vpp/22.06/developer/corearchitecture/vnet.html?highlight=pcap#pcap-rx-tx-and-drop-tracing
    https://s3-docs.fd.io/vpp/22.06/developer/corearchitecture/vnet.html?highlight=pcap#terse-description-of-the-mask-syntax

    (Note: documentation is outdated, therefore not all CLI calls can work in VPP 22.06)
    """

    output_file_prefix: str = "pcap-capture"
    """Prefix for packet capture output file. The resulting file will be in format 
    /tmp/<output_file_prefix>-<datetime-suffix>.pcap (the /tmp folder is fixed by VPP 22.06 implementation)"""

    max_packets_to_capture: int = 1000  # 1000 packets is VPP default
    """Packet-depth of local data buffer. Once the local buffer gets full, no more packets are captured. By removing 
    this whole configuration the packets capturing will be stopped and content of local buffer is flushed to file.
    
    Note: official VPP documentation tells that local buffer is flushed to file when it gets full, but that never 
    happens in VPP 22.02"""  # TODO Note still applies for VPP 22.10.1 that is used now ?
    max_bytes_per_packet: int = 9000
    """Maximum number of bytes to capture for each packet. Must be >= 32, <= 9000."""
    preallocate_data_buffer: bool = True
    """Preallocate the data buffer, to avoid vector expansion delays during pcap capture"""
    free_data_buffer: bool = False
    """Free the data buffer. Ordinarily it’s a feature to retain the data buffer so this option is seldom used."""

    # packet capture options (can use multiple of them at the same time to specify what packets you want to capture)

    capture_rx_packets: bool = True
    capture_tx_packets: bool = True
    capture_dropped_packets: bool = True
    capture_only_from_interface: str = PCAP_ANY_INTERFACE
    """Limiting capturing of packets to given interface (or using traffic from all interfaces if 
    PCAP_ANY_INTERFACE is used)"""
    capture_by_matching_rules: List[
        Union[L2PacketMatchingRule, L3PacketMatchingRule]
    ] = field(default_factory=lambda: [])
    """Match rules that will be applied for capturing packets. Only packets that match all rules(together with other 
    capturing restrictions from other RunningPacketCapture's fields) will be captured. You can combine multiple 
    L2/L3 rules."""  # (NOTE: There are L4 matching rules according to documentation, but they didn't work in VPP, giving error while defining any L4 rule)


@dataclass_json
@dataclass
class LinuxDockerVethTunnelConfigurationItem(ConfigurationItem):
    """Configuration for creating veth tunnel in linux (with one end in default linux namespace and other in docker
    container)"""

    docker_container_name: str = ""
    """Name of the docker container that should have one end of the veth tunnel"""
    host_interface_name: str = "host-veth"
    docker_interface_name: str = "veth"
    host_interface_ip_addresses: List[str] = field(default_factory=lambda: [])
    docker_interface_ip_addresses: List[str] = field(default_factory=lambda: [])


@dataclass_json
@dataclass
class LinuxRouteConfigurationItem(ConfigurationItem):
    """Configuration of static route in linux namespace (default linux namespace or docker container linux namespace)"""

    docker_container_name: str = ""
    """ Name of docker container into which you want to put linux route. Use empty string if the default linux network 
    namespace(linux host) should be used."""
    destination_network: str = ""
    """ Destination ip network for steering traffic using this route. 
    It has format <ip address>/<prefix> (i.e. "10.1.2.0/24)."""
    next_hop_address: str = ""
    """ Next hop ip address """
    outgoing_interface: str = ""
    """ Name of outgoing interface that should be used for this route. """


#################################
### API reply models
#################################


@dataclass_json
@dataclass
class ApplyReply:
    """Reply when applying configuration_item in handlers"""

    success: bool = False
    """Whether the appying was successfull or not"""
    error: str = ""
    """Error message if not success"""
    vpp_data: Dict[str, any] = field(default_factory=lambda: (dict()))
    """Additional data linked to the reply"""


@dataclass_json
@dataclass
class ConfigurationItemDetail:
    """ConfigurationItemRecord (internal configuration data structure) stripped of internal fields and ready for
    publishing to user through client API (get_configuration)"""

    # Note: normally ConfigurationItem should be used but that is prevented by serialization/deserialization
    # problems with combination of used libraries (Pyro5, dataclass-json) -> must list
    # all children of ConfigurationItem in Union
    config: Union[
        InterfaceConfigurationItem,
        RouteConfigurationItem,
        ACLConfigurationItem,
        Nat44AddressPoolConfigurationItem,
        Nat44InterfaceConfigurationItem,
        DNat44ConfigurationItem,
        WireguardPeerConfigurationItem,
        LCPGlobalsConfigurationItem,
        LCPPairConfigurationItem,
        FRRConfigurationItem,
        LinuxDockerVethTunnelConfigurationItem,
        LinuxRouteConfigurationItem,
        DHCPConfigurationItem,
    ]
    state: State
    metadata: Dict[str, Any] = field(default_factory=lambda: (dict()))


@dataclass_json
@dataclass
class ConfigurationItemReply(ConfigurationItemDetail):
    """ConfigurationItemDetail with added error from applying configuration item to VPP/FRR/Linux namespace"""

    error: str = ""


@dataclass_json
@dataclass
class AddConfigurationItemsReply:
    """Reply from applying configuration item using client.add_configuration"""

    processing_error: str = ""
    """Error that doesn't come from VPP/FRR/Linux namespace call, but is related to server processing of the request 
    (i.e. invalid input)"""
    all_added_items_applied_to_vpp: bool = True
    """Whether all configuration items given on client.add_configuration input were applied to VPP/FRR/Linux namespace 
    successfully (without any error). It is computed from added_items field."""
    added_items: List[ConfigurationItemDetail] = field(default_factory=lambda: [])
    """Server state of input configuration items after processing in server (see ConfigurationItemDetail). It can be 
    used to get current state of configuration items and more."""
    vpp_apply_success: bool = True
    """Whether all configuration in vpp_apply_attempted_items was applied or not. It is computed 
    from vpp_apply_attempted_items field."""
    vpp_apply_attempted_items: List[ConfigurationItemReply] = field(
        default_factory=lambda: []
    )
    """Server state of all configuration that can be applied (either was unblocked by satisfying configuration 
    dependencies or are new configuration items that are not blocked by dependency restriction). This configuration is 
    applied and you can use this inner server state (that is after applying) to check from success and more."""


@dataclass_json
@dataclass
class DeleteConfigurationItemsReply:
    """Reply from applying configuration item using client.delete_configuration"""

    processing_error: str = ""
    """Error that doesn't come from VPP/FRR/Linux namespace call, but is related to server processing of the request 
    (i.e. invalid input)"""
    vpp_apply_success: bool = True
    """Whether all configuration items given on client.delete_configuration input were applied(removal) 
    to VPP/FRR/Linux namespace successfully (without any error). It is computed from vpp_apply_attempted_items field."""
    vpp_apply_attempted_items: List[ConfigurationItemReply] = field(
        default_factory=lambda: []
    )
    """Server state of input configuration items after processing in server (see ConfigurationItemDetail). It can be 
    used to get current state of configuration items and more. These information (in case of successfully removal) are 
    removed from server and will be missing from retrieved data from get_configuration() call."""


@dataclass_json
@dataclass
class UpdateConfigurationItemsReply:
    """Reply from applying configuration item using client.update_configuration"""

    processing_error: str = ""
    """Error that doesn't come from VPP/FRR/Linux namespace call, but is related to server processing of the request 
    (i.e. invalid input)"""
    all_updated_items_applied_to_vpp: bool = True
    """Similar to AddConfigurationItemsReply.all_added_items_applied_to_vpp"""
    updated_items: List[ConfigurationItemDetail] = field(default_factory=lambda: [])
    """Similar to AddConfigurationItemsReply.added_items"""
    vpp_apply_success: bool = True
    """Similar to (AddConfigurationItemsReply.vpp_apply_success and DeleteConfigurationItemsReply.vpp_apply_success)"""
    vpp_apply_attempted_items: List[ConfigurationItemReply] = field(
        default_factory=lambda: []
    )
    """Similar to (AddConfigurationItemsReply.vpp_apply_attempted_items and 
    DeleteConfigurationItemsReply.vpp_apply_attempted_items)"""


@dataclass_json
@dataclass
class GetConfigurationItemsReply:
    """Reply from applying configuration item using client.get_configuration"""

    items: List[ConfigurationItemDetail] = field(default_factory=lambda: [])
    """Server state of all configuration items that server currently knows."""


@dataclass_json
@dataclass
class ProcessCallReply:
    """Reply that transport subprocess run result"""

    return_code: int = 0
    """subprocess run result code"""
    stdout: str = ""
    """subprocess run's stdout"""
    stderr: str = ""
    """subprocess run's stderr"""


#################################
### Dump models
#################################


@dataclass_json
@dataclass
class SWInterfaceDetail:
    b_dmac: str = ""
    b_smac: str = ""
    b_vlanid: int = 0
    context: int = 0
    flags: int = 0
    i_sid: int = 0
    interface_dev_type: str = ""
    interface_name: str = ""
    l2_address: str = ""
    link_duplex: int = 0  # i.e. <vl_api_link_duplex_t.LINK_DUPLEX_API_UNKNOWN: 0>
    link_mtu: int = 0
    link_speed: int = 0
    mtu: list[int] = field(default_factory=lambda: [])
    outer_tag: int = 0
    sub_id: int = 0
    sub_if_flags: int = 0
    sub_inner_vlan_id: int = 0
    sub_number_of_tags: int = 0
    sub_outer_vlan_id: int = 0
    sup_sw_if_index: int = 0
    sw_if_index: int = 0
    tag: str = ""
    type: int = 0  # i.e. <vl_api_if_type_t.IF_API_TYPE_HARDWARE: 0>
    vtr_op: int = 0
    vtr_push_dot1q: int = 0
    vtr_tag1: int = 0
    vtr_tag2: int = 0

    def is_link_state_up(self):
        return True if self.flags & IF_STATUS_API_FLAG_LINK_UP else False

    def is_admin_state_up(self):
        return True if self.flags & IF_STATUS_API_FLAG_ADMIN_UP else False

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return SWInterfaceDetail(
            b_dmac=obj.b_dmac,
            b_smac=obj.b_smac,
            b_vlanid=obj.b_vlanid,
            context=obj.context,
            i_sid=obj.i_sid,
            interface_dev_type=obj.interface_dev_type,
            interface_name=obj.interface_name,
            l2_address=obj.l2_address,
            link_mtu=obj.link_mtu,
            link_speed=obj.link_speed,
            outer_tag=obj.outer_tag,
            sub_id=obj.sub_id,
            sub_inner_vlan_id=obj.sub_inner_vlan_id,
            sub_number_of_tags=obj.sub_number_of_tags,
            sub_outer_vlan_id=obj.sub_outer_vlan_id,
            sup_sw_if_index=obj.sup_sw_if_index,
            sw_if_index=obj.sw_if_index,
            tag=obj.tag,
            vtr_op=obj.vtr_op,
            vtr_push_dot1q=obj.vtr_push_dot1q,
            vtr_tag1=obj.vtr_tag1,
            vtr_tag2=obj.vtr_tag2,
            link_duplex=int(
                obj.link_duplex
            ),  # i.e. <vl_api_link_duplex_t.LINK_DUPLEX_API_UNKNOWN: 0>
            type=int(obj.type),  # i.e. <vl_api_if_type_t.IF_API_TYPE_HARDWARE: 0>
            flags=int(obj.flags),  # typing.Set[VppEnum.vl_api_if_status_flags_t]
            mtu=obj.mtu,
            sub_if_flags=int(
                obj.sub_if_flags
            ),  # typing.Set[VppEnum.vl_api_sub_if_flags_t]
        )


@dataclass_json
@dataclass
class FibPathNextHop:  # vl_api_fib_path_nh_t
    ipv4_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    ipv6_address: ipaddress.IPv6Address = field(
        default_factory=lambda: ipaddress.IPv6Address("::"),
        metadata=config(mm_field=fields.IPv6()),
    )

    classify_table_index: int = 0
    obj_id: int = 0
    via_label: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return FibPathNextHop(
            ipv4_address=obj.address.ip4,
            ipv6_address=obj.address.ip6,
            classify_table_index=obj.classify_table_index,
            obj_id=obj.obj_id,
            via_label=obj.via_label,
        )


@dataclass_json
@dataclass
class FibPath:  # vl_api_fib_path_t
    flags: int = 0  # vl_api_fib_path_flags_t
    # label_stack: List[vl_api_fib_mpls_label_t]
    # n_labels: int = 0
    nh: typing.Union[FibPathNextHop, None] = None
    preference: int = 0
    proto: int = 0  # vl_api_fib_path_nh_proto_t
    rpf_id: int = 0
    sw_if_index: int = 4294967295
    table_id: int = 0
    type: int = 0  # vl_api_fib_path_type_t
    weight: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return FibPath(
            flags=int(obj.flags),
            nh=(
                FibPathNextHop.get_from_basic_vpp_client_equivalent_obj(obj.nh)
                if obj.nh
                else None
            ),
            preference=int(obj.preference),
            proto=int(obj.proto),
            rpf_id=int(obj.rpf_id),
            sw_if_index=int(obj.sw_if_index),
            table_id=int(obj.table_id),
            type=int(obj.type),
            weight=int(obj.weight),
        )


@dataclass_json
@dataclass
class IPRouteDetail:  # ignoring context, using only route field of type vl_api_ip_route_t
    n_paths: int = 0
    paths: List[FibPath] = field(default_factory=lambda: [])
    prefix: ipaddress.IPv4Network = field(
        default_factory=lambda: ipaddress.IPv4Network("0.0.0.0/0"),
        metadata=config(
            mm_field=custom_marshmallow_fields.IPv4Network()  # missing in builtin marshmallow fields
        ),
    )
    stats_index: int = 0
    table_id: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return IPRouteDetail(
            n_paths=obj.route.n_paths,
            paths=[
                FibPath.get_from_basic_vpp_client_equivalent_obj(path)
                for path in obj.route.paths
            ],
            prefix=obj.route.prefix,
            stats_index=obj.route.stats_index,
            table_id=obj.route.table_id,
        )


@dataclass_json
@dataclass
class ACLRule:  # vl_api_acl_rule_t
    is_permit: int = 0  # vl_api_acl_action_t
    dst_prefix: ipaddress.IPv4Network = ipaddress.IPv4Network("0.0.0.0/0")
    src_prefix: ipaddress.IPv4Network = ipaddress.IPv4Network("0.0.0.0/0")
    dstport_or_icmpcode_first: int = 0
    dstport_or_icmpcode_last: int = 0
    srcport_or_icmptype_first: int = 0
    srcport_or_icmptype_last: int = 0
    proto: int = 0  # vl_api_ip_proto_t
    tcp_flags_mask: int = 0
    tcp_flags_value: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return ACLRule(
            is_permit=obj.is_permit,
            dst_prefix=obj.dst_prefix,
            src_prefix=obj.src_prefix,
            dstport_or_icmpcode_first=obj.dstport_or_icmpcode_first,
            dstport_or_icmpcode_last=obj.dstport_or_icmpcode_last,
            srcport_or_icmptype_first=obj.srcport_or_icmptype_first,
            srcport_or_icmptype_last=obj.srcport_or_icmptype_last,
            proto=obj.proto,
            tcp_flags_mask=obj.tcp_flags_mask,
            tcp_flags_value=obj.tcp_flags_value,
        )


@dataclass_json
@dataclass
class ACLDetails:  # acl_details
    acl_index: int = 0
    tag: str = ""
    count: int = 0  # rules count
    r: List[ACLRule] = field(default_factory=lambda: [])  # rules

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return ACLDetails(
            acl_index=obj.acl_index,
            tag=obj.tag,
            count=obj.count,
            r=[
                ACLRule.get_from_basic_vpp_client_equivalent_obj(rule) for rule in obj.r
            ],
        )


@dataclass_json
@dataclass
class NAT44AddressDetails:  # nat44_address_details
    flags: int = 0  # vl_api_nat_config_flags_t
    ip_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )

    # vrf_id: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return NAT44AddressDetails(
            flags=int(obj.flags),
            ip_address=obj.ip_address,
        )


@dataclass_json
@dataclass
class NAT44InterfaceDetails:  # nat44_interface_details
    flags: int = 0  # vl_api_nat_config_flags_t
    sw_if_index: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return NAT44InterfaceDetails(
            flags=int(obj.flags),
            sw_if_index=obj.sw_if_index,
        )


@dataclass_json
@dataclass
class NAT44StaticMappingDetails:  # nat44_static_mapping_details
    external_sw_if_index: int = 0
    external_ip_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    external_port: int = 0
    local_ip_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    local_port: int = 0
    protocol: int = 0
    tag: str = ""
    flags: int = 0  # vl_api_nat_config_flags_t

    # vrf_id: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return NAT44StaticMappingDetails(
            external_sw_if_index=obj.external_sw_if_index,
            external_ip_address=obj.external_ip_address,
            external_port=obj.external_port,
            local_ip_address=obj.local_ip_address,
            local_port=obj.local_port,
            protocol=obj.protocol,
            tag=obj.tag,
            flags=obj.flags,
        )


@dataclass_json
@dataclass
class NAT44IdentityMappingDetails:  # nat44_identity_mapping_details
    sw_if_index: int = 0
    ip_address: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    port: int = 0
    protocol: int = 0
    tag: str = ""
    flags: int = 0  # vl_api_nat_config_flags_t

    # vrf_id: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return NAT44IdentityMappingDetails(
            sw_if_index=obj.sw_if_index,
            ip_address=obj.ip_address,
            port=obj.port,
            protocol=obj.protocol,
            tag=obj.tag,
            flags=obj.flags,
        )


@dataclass_json
@dataclass
class WireguardPeerDetail:  # vl_api_wireguard_peer_t (=peer attribute from dump returning wireguard_peers_details)
    endpoint: ipaddress.IPv4Address = field(
        default_factory=lambda: ipaddress.IPv4Address("0.0.0.0"),
        metadata=config(mm_field=fields.IPv4()),
    )
    port: int = 0
    n_allowed_ips: int = 0
    allowed_ips: List[ipaddress.IPv4Network] = field(default_factory=lambda: [])
    public_key: str = ""
    """Base64 encoded public key"""
    sw_if_index: int = 0
    persistent_keepalive: int = 0
    flags: int = 0  # vl_api_wireguard_peer_flags_t
    table_id: int = 0

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return WireguardPeerDetail(
            endpoint=obj.endpoint,
            port=obj.port,
            n_allowed_ips=obj.n_allowed_ips,
            allowed_ips=obj.allowed_ips,
            public_key=base64.b64encode(obj.public_key).decode(),
            sw_if_index=obj.sw_if_index,
            persistent_keepalive=obj.persistent_keepalive,
            flags=int(obj.flags),
            table_id=obj.table_id,
        )


@dataclass_json
@dataclass
class LCPPairDetail:  # lcp_itf_pair_details
    phy_sw_if_index: int = 0
    host_sw_if_index: int = 0
    vif_index: int = (0,)
    host_if_name: str = ""
    host_if_type: int = 0  # vl_api_lcp_itf_host_type_t
    netns: str = ""

    @staticmethod
    def get_from_basic_vpp_client_equivalent_obj(obj):
        """Static conversion of non-persistable result of VPP dump call into modeled and persistable version
        of the same data"""

        return LCPPairDetail(
            phy_sw_if_index=obj.phy_sw_if_index,
            host_sw_if_index=obj.host_sw_if_index,
            vif_index=obj.vif_index,
            host_if_name=obj.host_if_name,
            host_if_type=int(obj.host_if_type),  # vl_api_lcp_itf_host_type_t
            netns=obj.netns,
        )
